# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..helpers.rendering import SaganSet


class ValueKeyAggregator(object):
    """Aggregator based on tha actual value of the key/attribute"""
    def __init__(self, key, prefix=None):
        self.aggregation_keys = key.split('.')
        self.key_prefix = prefix or self.aggregation_keys[-1].upper()

    def get_key_value(self, entity):
        """
        Returns the value of the key/attribute the aggregation will use to
        bucketize probes/results
        """
        attribute = entity
        for key in self.aggregation_keys:
            attribute = getattr(attribute, key)
        return attribute

    def get_bucket(self, entity):
        """
        Returns the bucket the specific entity belongs to based on the give
        key/attribute
        """
        return "{0}: {1}".format(self.key_prefix, self.get_key_value(entity))

    def insert2bucket(self, buckets, bucket, entity):
        if bucket in buckets:
            buckets[bucket].append(entity)
        else:
            buckets[bucket] = [entity]


class RangeKeyAggregator(ValueKeyAggregator):
    """
    Aggregator based on where the position of the value of the key/attribute is
    in the given range
    """

    def __init__(self, key, ranges):
        ValueKeyAggregator.__init__(self, key)
        self.aggregation_ranges = sorted(ranges, reverse=True)

    def get_bucket(self, entity):
        """
        Returns the bucket the specific entity belongs to based on the give
        key/attribute
        """

        bucket = "{0}: < {1}".format(
            self.key_prefix, self.aggregation_ranges[-1])

        key_value = self.get_key_value(entity)
        for index, krange in enumerate(self.aggregation_ranges):
            if key_value > krange:
                if index == 0:
                    bucket = "{0}: > {1}".format(self.key_prefix, krange)
                else:
                    bucket = "{0}: {1}-{2}".format(
                        self.key_prefix,
                        krange,
                        self.aggregation_ranges[index - 1]
                    )
                break

        return bucket


def aggregate(entities, aggregators):
    """
    This is doing the len(aggregators) level aggregation of the entities.
    Caution: being recursive is a bit hard to read/understand, if you change
    something make sure you run tests.
    """

    if not aggregators:
        return entities

    if isinstance(entities, (list, SaganSet)):

        aggregator = aggregators.pop(0)
        buckets = {}
        for entity in entities:
            bucket = aggregator.get_bucket(entity)
            aggregator.insert2bucket(buckets, bucket, entity)
        return aggregate(buckets, aggregators)

    elif isinstance(entities, dict):

        for k, v in entities.items():
            entities[k] = aggregate(entities[k], aggregators[:])

    return entities



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import OpenSSL

from ..helpers.sanitisers import sanitise
from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_SSLCERT]

    def on_result(self, result):
        r = ""
        for certificate in result.certificates:
            r += self.get_formatted_response(certificate)
        return "\nProbe #{0}\n{1}\n".format(result.probe_id, r)

    @classmethod
    def get_formatted_response(cls, certificate):
        x509 = OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM,
            certificate.raw_data.replace("\\/", "/").replace("\n\n", "\n")
        )

        pkey_type = x509.get_pubkey().type()

        # TODO: to be improved
        if pkey_type == 6:
            pkey_type_descr = "rsaEncryption"
        else:
            pkey_type_descr = pkey_type

        return cls.render(
            "reports/sslcert.txt",
            issuer_c=sanitise(certificate.issuer_c),
            issuer_o=sanitise(certificate.issuer_o),
            issuer_cn=sanitise(certificate.issuer_cn),
            not_before=certificate.valid_from,
            not_after=certificate.valid_until,
            subject_c=sanitise(certificate.subject_c),
            subject_o=sanitise(certificate.subject_o),
            subject_cn=sanitise(certificate.subject_cn),
            version=x509.get_version(),
            serial_number=x509.get_serial_number(),
            signature_algorithm=x509.get_signature_algorithm(),
            pkey_type=pkey_type_descr,
            pkey_bits=x509.get_pubkey().bits(),
            sha1fp=certificate.checksum_sha1,
            sha256fp=certificate.checksum_sha256
        )



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime
import unittest

try:
    from unittest import mock  # Python 3.4+
except ImportError:
    import mock

from ripe.atlas.tools.commands.measurement_search import Command

from ..base import capture_sys_output

COMMAND_MODULE = "ripe.atlas.tools.commands.measurement_search"


class FakeGen(object):
    """
    A rip-off of the code used for testing probes, but a little prettier.
    """

    Measurement = collections.namedtuple("Measurement", (
        "id", "type", "status", "status_id", "meta_data", "target",
        "description"
    ))

    def __init__(self):
        self.data = [
            self.Measurement(
                id=1, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 1", description="Description 1",
            ),
            self.Measurement(
                id=2, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 2", description="Description 2",
            ),
            self.Measurement(
                id=3, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 3", description="Description 3",
            ),
            self.Measurement(
                id=4, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 4", description="Description 4",
            ),
            self.Measurement(
                id=5, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 5", description="Description 5",
            ),
        ]
        self.total_count = 5

    def __iter__(self):
        return self

    # Python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        if not self.data:
            raise StopIteration()
        else:
            return self.data.pop(0)


class TestMeasurementsCommand(unittest.TestCase):

    @mock.patch("{}.MeasurementRequest".format(COMMAND_MODULE))
    def test_with_empty_args(self, mock_request):

        mock_request.return_value = FakeGen()

        cmd = Command()
        with capture_sys_output() as (stdout, stderr):
            cmd.init_args([])
            cmd.run()

        expected_content = [
            "",
            "Id      Type       Description                                            Status",
            "================================================================================",
            str(b"1       ping       Description 1                                         Ongoing"),
            str(b"2       ping       Description 2                                         Ongoing"),
            str(b"3       ping       Description 3                                         Ongoing"),
            str(b"4       ping       Description 4                                         Ongoing"),
            str(b"5       ping       Description 5                                         Ongoing"),
            "================================================================================",
            "                                               Showing 5 of 5 total measurements",
            "",
        ]
        self.maxDiff = None
        self.assertEqual(
            set(stdout.getvalue().split("\n")),
            set(expected_content)
        )
        self.assertEqual(
            cmd.arguments.field, ("id", "type", "description", "status"))

    @mock.patch("{}.MeasurementRequest".format(COMMAND_MODULE))
    def test_get_line_items(self, mock_request):

        mock_request.return_value = FakeGen()
        cmd = Command()
        cmd.init_args([])
        cmd.run()
        self.assertEqual(
            cmd._get_line_items(FakeGen.Measurement(
                id=1, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 1", description="Description 1",
            )),
            [1, "ping", "Description 1", "Ongoing"]
        )

        cmd = Command()
        cmd.init_args([
            "--field", "id",
            "--field", "status"
        ])
        self.assertEqual(
            cmd._get_line_items(FakeGen.Measurement(
                id=1, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 1", description="Description 1",
            )),
            [1, "Ongoing"]
        )

        cmd = Command()
        cmd.init_args([
            "--field", "url",
        ])
        self.assertEqual(
            cmd._get_line_items(FakeGen.Measurement(
                id=1, type="ping", status="Ongoing", status_id=2,
                meta_data={"status": {"name": "Ongoing", "id": 2}},
                target="Name 1", description="Description 1",
            )),
            ["https://atlas.ripe.net/measurements/1/"]
        )

    def test_get_filters(self):
        cmd = Command()
        cmd.init_args([
            "--search", "the force is strong with this one",
            "--status", "ongoing",
            "--af", "6",
            "--type", "ping",
            "--started-before", "2015-01-01",
            "--started-after", "2014-01-01",
            "--stopped-before", "2015-01-01",
            "--stopped-after", "2014-01-01",
        ])
        self.assertEqual(cmd._get_filters(), {
            "search": "the force is strong with this one",
            "status__in": (2,),
            "af": 6,
            "type": "ping",
            "start_time__lt": datetime.datetime(2015, 1, 1),
            "start_time__gt": datetime.datetime(2014, 1, 1),
            "stop_time__lt": datetime.datetime(2015, 1, 1),
            "stop_time__gt": datetime.datetime(2014, 1, 1),
        })

    def test_get_colour_from_status(self):
        cmd = Command()
        self.assertEqual(cmd._get_colour_from_status(0), "blue")
        self.assertEqual(cmd._get_colour_from_status(1), "blue")
        self.assertEqual(cmd._get_colour_from_status(2), "green")
        self.assertEqual(cmd._get_colour_from_status(4), "yellow")
        self.assertEqual(cmd._get_colour_from_status(5), "red")
        self.assertEqual(cmd._get_colour_from_status(6), "red")
        self.assertEqual(cmd._get_colour_from_status(7), "red")
        self.assertEqual(cmd._get_colour_from_status("XXX"), "white")

    def test_fail_arguments(self):
        expected_failures = (
            ("--status", "not a status"),
            ("--not-an-option",),
            ("--af", "5"),
            ("--type", "not a type"),
            ("--field", "not a field"),
        )
        for failure in expected_failures:
            with capture_sys_output():
                with self.assertRaises(SystemExit):
                    Command().init_args(failure)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ...helpers.validators import ArgumentType
from ...settings import conf
from .base import Command


class PingMeasureCommand(Command):

    def add_arguments(self):

        Command.add_arguments(self)

        spec = conf["specification"]["types"]["ping"]

        specific = self.parser.add_argument_group("Ping-specific Options")
        specific.add_argument(
            "--packets",
            type=ArgumentType.integer_range(minimum=1, maximum=16),
            default=spec["packets"],
            help="The number of packets sent"
        )
        specific.add_argument(
            "--size",
            type=ArgumentType.integer_range(minimum=1, maximum=2048),
            default=spec["size"],
            help="The size of packets sent"
        )
        specific.add_argument(
            "--packet-interval",
            type=ArgumentType.integer_range(minimum=2, maximum=30000),
            default=spec["packet-interval"],
        )
        specific.add_argument(
            "--include-probe-id",
            default=spec["include_probe_id"],
            action="store_true",
            help="Include the ASCII-encoded probe ID in the ping packets",
        )

    def _get_measurement_kwargs(self):

        r = Command._get_measurement_kwargs(self)

        r["packets"] = self.arguments.packets
        r["packet_interval"] = self.arguments.packet_interval
        r["size"] = self.arguments.size
        if self.arguments.include_probe_id:
            r["include_probe_id"] = True

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import unittest

# Python 3.4+ comes with mock in unittest
try:
    from unittest import mock
except ImportError:
    import mock

from ripe.atlas.tools.commands.alias import Command
from ripe.atlas.tools.exceptions import RipeAtlasToolsException
from ripe.atlas.tools.settings import AliasesDB

from ..base import capture_sys_output


class FakeAliasesDB(AliasesDB):

    @staticmethod
    def write(aliases):
        pass

class TestAliasCommand(unittest.TestCase):

    ALIASES_PATH = "ripe.atlas.tools.commands.alias.aliases"
    ALIASES = {
        "measurement": {
            "msm1": 1
        },
        "probe": {
            "prb1": 1
        }
    }
    ALIASES_CLASS_PATH = "ripe.atlas.tools.commands.alias.AliasesDB"

    def setUp(self):
        self.cmd = Command()
        self.aliases = copy.deepcopy(TestAliasCommand.ALIASES)
        mock.patch(self.ALIASES_CLASS_PATH, FakeAliasesDB).start()

    def tearDown(self):
        mock.patch.stopall()

    def test_no_arguments(self):
        with capture_sys_output():
            with self.assertRaises(RipeAtlasToolsException) as e:
                self.cmd.init_args([])
                self.cmd.run()
            self.assertTrue(
                str(e.exception).startswith("Action not given."))

    def test_bad_action(self):
        with capture_sys_output():
            with self.assertRaises(SystemExit) as e:
                self.cmd.init_args(["test"])

    def test_bad_target_id(self):
        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit) as e:
                self.cmd.init_args("add probe a b".split())
            err = stderr.getvalue().split("\n")[-2]
        self.assertEqual(
            err,
            "Ripe-atlas alias add: error: argument target: invalid int value: 'a'"
        )

    def test_add_args(self):
        for alias_type in ("probe", "measurement"):
            with capture_sys_output() as (stdout, stderr):
                with self.assertRaises(SystemExit) as e:
                    cmd = Command()
                    cmd.init_args(["add", alias_type])

                with self.assertRaises(SystemExit) as e:
                    cmd = Command()
                    cmd.init_args(["add", alias_type, "1"])

                with self.assertRaises(SystemExit) as e:
                    cmd = Command()
                    cmd.init_args(["add", alias_type, "1", "1"])

                try:
                    cmd = Command()
                    cmd.init_args(["add", alias_type, "1", "one"])
                except Exception as e:
                    self.fail("Failed with {}".format(str(e)))

    def test_del_args(self):
        for alias_type in ("probe", "measurement"):
            with capture_sys_output() as (stdout, stderr):
                with self.assertRaises(SystemExit) as e:
                    cmd = Command()
                    cmd.init_args(["del", alias_type])

                with self.assertRaises(SystemExit) as e:
                    cmd = Command()
                    cmd.init_args(["del", alias_type, "1"])

                try:
                    cmd = Command()
                    cmd.init_args(["del", alias_type, "one"])
                except Exception as e:
                    self.fail("Failed with {}".format(str(e)))

    def test_bad_alias(self):
        with capture_sys_output() as (stdout, stderr):
            with self.assertRaises(SystemExit) as e:
                self.cmd.init_args("add measurement 1 bad+alias".split())
            err = stderr.getvalue().split("\n")[-2]
        self.assertEqual(
            err,
            'Ripe-atlas alias add: error: argument alias: '
            '"bad+alias" does not appear to be a valid alias.'
        )

    def test_show_msm_ok(self):
        path = "ripe.atlas.tools.commands.alias.Command.ok"
        with mock.patch(path) as mock_ok:
            with mock.patch(self.ALIASES_PATH, self.aliases):
                self.cmd.init_args("show measurement msm1".split())
                self.cmd.run()
        mock_ok.assert_called_once()

    def test_show_msm_ko(self):
        path = "ripe.atlas.tools.commands.alias.Command.not_ok"
        with mock.patch(path) as mock_ko:
            with mock.patch(self.ALIASES_PATH, self.aliases):
                self.cmd.init_args("show measurement msm2".split())
                self.cmd.run()
        mock_ko.assert_called_once()

    def test_add_msm(self):
        with mock.patch(self.ALIASES_PATH, self.aliases):
            self.cmd.init_args("add measurement 2 msm2".split())
            self.cmd.run()
        self.assertTrue("msm2" in self.aliases["measurement"])
        self.assertEqual(
            self.aliases["measurement"]["msm2"],
            2
        )

    def test_del_msm(self):
        self.assertTrue("msm1" in self.aliases["measurement"])
        with mock.patch(self.ALIASES_PATH, self.aliases):
            self.cmd.init_args("del measurement msm1".split())
            self.cmd.run()
        self.assertFalse("msm1" in self.aliases["measurement"])

    def test_list(self):
        path = "ripe.atlas.tools.commands.alias.Command.ok"
        with mock.patch(path) as mock_ok:
            self.aliases["measurement"]["msm2"] = 2
            self.aliases["measurement"]["abc"] = 123
            with mock.patch(self.ALIASES_PATH, self.aliases):
                self.cmd.init_args("list measurement".split())
                self.cmd.run()
        mock_ok.assert_called_once_with(
            "Measurement aliases:\n\n- abc: 123\n- msm1: 1\n- msm2: 2\n"
        )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
import argparse
import importlib
import os
import pkgutil
import re
import six
import sys

from ..helpers.colours import colourise
from ..version import __version__


class RipeHelpFormatter(argparse.RawTextHelpFormatter):

    def _format_usage(self, *args):
        r = argparse.RawTextHelpFormatter._format_usage(
            self, *args).capitalize()
        return "\n\n{}\n".format(r)


def _get_command_name(cmd_or_factory):
    return cmd_or_factory.NAME if cmd_or_factory.NAME else \
        cmd_or_factory.__module__.rsplit(".", 1)[-1]


class Command(object):
    # If NAME isn't defined then the module name is used
    # It exists basically to allow hyphens in the command names
    NAME = None
    DESCRIPTION = ""  # Define this in the subclass
    _commands = None

    DEPRECATED_ALIASES = {
        "measurement": "measurement-info",
        "measurements": "measurement-search",
        "probe": "probe-info",
        "probes": "probe-search",
        "render": "report",
    }

    def __init__(self, *args, **kwargs):

        self.arguments = None
        self.parser = argparse.ArgumentParser(
            formatter_class=RipeHelpFormatter,
            description="\n\n".join([
                self.DESCRIPTION + ".",
                getattr(self, "EXTRA_DESCRIPTION", ""),
            ]),
            prog="ripe-atlas {}".format(self.get_name())
        )
        self.user_agent = self._get_user_agent()

    @classmethod
    def get_name(cls):
        return _get_command_name(cls)

    @staticmethod
    def _get_packages_for_paths(paths):
        """
        Yield path, package_name for all of the packages found in `paths`.
        """
        for loader, package_name, _ in pkgutil.iter_modules(paths):
            yield loader.path, package_name

    @classmethod
    def _get_user_command_path(cls):
        user_base_path = os.path.join(
            os.path.expanduser("~"), ".config", "ripe-atlas-tools",
        )
        return os.path.join(user_base_path, "commands")

    @classmethod
    def _load_commands(cls):
        """
        Scan for available commands and store a map of command names to module
        paths.
        """
        builtin_path = os.path.dirname(__file__)
        user_command_path = cls._get_user_command_path()

        cls._commands = {}

        paths = [builtin_path, user_command_path]
        for path, package_name in cls._get_packages_for_paths(paths):
            if package_name == "base":
                continue
            if path == builtin_path:
                module = "ripe.atlas.tools.commands.{}".format(package_name)
            else:
                module = package_name
                if user_command_path not in sys.path:
                    sys.path.append(user_command_path)
            cls._commands[package_name] = module

    @classmethod
    def load_command_class(cls, command_name):
        """
        Get the Command or Factory with the given command name.
        """
        if command_name in cls.DEPRECATED_ALIASES:
            alias = command_name
            command_name = cls.DEPRECATED_ALIASES[alias]
            sys.stderr.write(colourise(
                "Warning: {} is a deprecated alias for {}\n\n".format(
                    alias, command_name,
                ),
                "yellow"
            ))

        if cls._commands is None:
            cls._load_commands()

        try:
            module_name = cls._commands[command_name.replace("-", "_")]
        except KeyError:
            return

        module = importlib.import_module(module_name)

        if hasattr(module, "Factory"):
            cmd = module.Factory
        else:
            cmd = module.Command

        return cmd

    @classmethod
    def get_available_commands(cls):
        """
        Get a list of commands that we can execute.  By default, we have a
        fixed list that we make available in this directory, but the user can
        create her own plugins and store them at
        ~/.config/ripe-atlas-tools/commands/.  If we find any files there, we
        add them to the list here.
        """
        if not cls._commands:
            cls._load_commands()

        return sorted(cls._commands.keys())

    def init_args(self, args=None):
        """
        Initialises all parse arguments and makes them available to the class.
        """

        if args is None:
            args = sys.argv[1:]

        self.arguments = self.parser.parse_args(
            self._modify_parser_args(args))

    def run(self):
        raise NotImplemented()

    def add_arguments(self):
        """

        A hook that's executed in the __init__, you can make use of
        `self.parser` here to add arguments to the command:

          self.parser.add_argument(
            "measurement_id",
            type=int,
            help="The measurement id you want to use"
          )

        """
        pass

    def _modify_parser_args(self, args):
        """
        A modifier hook that can be overridden in the child class to allow that
        class to manipulate the arguments before being parsed.  The common
        use-case we're trying to solve here is popping a secondary argument off
        of the list and/or appending `--help` in some circumstances.
        """

        self.add_arguments()

        return args

    def ok(self, message):
        if sys.stdout.isatty():
            sys.stdout.write("\n{}\n\n".format(colourise(message, "green")))

    def not_ok(self, message):
        if sys.stdout.isatty():
            sys.stdout.write("\n{}\n\n".format(colourise(message, "red")))

    @staticmethod
    def _get_user_agent():
        """
        Allow packagers to change the user-agent to whatever they like by
        placing a file called `user-agent` into the `tools` directory.  If no
        file is found, we go with a sensible default + the version.
        """

        try:
            custom = os.path.join(os.path.dirname(__file__), "..", "user-agent")
            with open(custom) as f:
                return f.readline().strip()[:128]
        except IOError:
            pass  # We go with the default

        return "RIPE Atlas Tools (Magellan) {}".format(__version__)


class TabularFieldsMixin(object):
    """
    A handy mixin to dump into classes that are expected to render tabular data.
    It expects both that COLUMNS is defined by the subclass and that --field is
    set in the add_arguments() method.
    """

    def _get_line_format(self):
        """
        Loop over the field arguments and generate a string that makes use of
        Python's string format mini language.  We later use this string to
        format the values for each row.
        """
        r = u""
        for field in self.arguments.field:
            if r:
                r += u" "
            r += (u"{!s:" + u"{}{}".format(*self.COLUMNS[field]) + u"}")
        return r

    def _get_header_names(self):
        return [_.capitalize() for _ in self.arguments.field]

    def _get_header(self):
        """
        Generates a header by using the line formatter and the list of field
        arguments.
        """
        return self._get_line_format().format(*self._get_header_names())

    def _get_horizontal_rule(self):
        """
        A bit of a hack: We get a formatted line for no other reason than to
        determine the width of that line.  Then we use a regex to overwrite that
        line with "=".
        """
        return re.sub(
            r".", "=", self._get_line_format().format(*self.arguments.field))

    def _get_line_items(self, measurement):
        raise NotImplementedError("This needs to be defined in the subclass.")

    def _get_filter_display(self, filters):

        if not filters:
            return ""

        r = colourise("\nFilters:\n", "white")
        for k, v in filters.items():
            if k not in ("search",):
                v = str(v).capitalize()
            r += colourise(
                "  {}: {}\n".format(*self._get_filter_key_value_pair(k, v)),
                "cyan"
            )

        return r

    def _get_filter_key_value_pair(self, k, v):
        return k.capitalize().replace("__", " "), v


class MetaDataMixin(object):

    @staticmethod
    def _prettify_boolean(boolean):

        checkmark = u"\u2714"
        x = u"\u2718"

        if boolean:
            return colourise(checkmark, "green")
        return colourise(x, "red")

    @staticmethod
    def _prettify_time(dtime):
        if isinstance(dtime, datetime):
            return "{} UTC".format(
                dtime.isoformat().replace("T", " "))

        return str(dtime)

    @staticmethod
    def _render_line(header, value):
        # Make sure we don't mix unicode and strings
        if six.PY2 and isinstance(value, six.string_types):
            value = unicode(value)

        log = u"{}  {}".format(colourise("{:25}".format(header), "bold"), value).encode("utf-8")
        print(log)


class Factory(object):
    NAME = None

    @classmethod
    def build(cls, *args, **kwargs):
        return object()

    @classmethod
    def get_name(cls):
        return _get_command_name(cls)



import os
import unittest
import subprocess


class BashCompletionTests(unittest.TestCase):
    """
    Testing the Python level bash completion code.
    This requires setting up the environment as if we got passed data
    from bash.
    """

    def setUp(self):
        os.environ['RIPE_ATLAS_AUTO_COMPLETE'] = '1'

    def _setup_env(self, substring):
        input_str = "ripe-atlas" + substring
        os.environ['COMP_WORDS'] = input_str
        comp_cword = len(input_str.split(' ')) - 1  # Index of the last word
        os.environ['COMP_CWORD'] = str(comp_cword)

    def _autocomplete(self, substring):

        self._setup_env(substring)
        cmd_parts = "ripe-atlas" + substring
        envs = os.environ.copy()
        process = subprocess.Popen(
            cmd_parts, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=envs, shell=True
        )
        output, error = process.communicate()
        return output.decode("utf-8"), error.decode("utf-8")

    def test_commands_completion(self):
        """Tests autocompletion of commands."""
        input_str = " "
        output, error = self._autocomplete(input_str)
        print(output, error)
        self.assertTrue("report" in output)

    def test_completion_disable(self):
        """
        Tests if autocompletion is disabled if environmental variable
        is not set.
        """
        input_str = " mea"
        del os.environ['RIPE_ATLAS_AUTO_COMPLETE']
        output, error = self._autocomplete(input_str)
        print(output, error)
        self.assertTrue("No such command" in error)

    def test_command_completion(self):
        """Tests autocompletion of specific command."""
        input_str = " meas"
        output, error = self._autocomplete(input_str)
        print(output, error)
        self.assertTrue("measure" in output)

    def test_options_completion(self):
        """Tests autocompletion of existing options for a command."""
        input_str = " measure "
        output, error = self._autocomplete(input_str)
        print(output, error)
        self.assertEqual(output, "dns http ntp ping sslcert traceroute")

    def test_option_completion(self):
        """Tests autocompletion of specific option of a command."""
        input_str = " measure ping --h"
        output, error = self._autocomplete(input_str)
        print(output, error)
        self.assertEqual(output, "--help")



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .aggregators import TestAggregators
from .commands import (
    TestProbesCommand,
    TestMeasureCommand,
    TestMeasurementsCommand,
    TestReportCommand,
    TestCommandLoading,
)
from .helpers import TestArgumentTypeHelper
from .renderers import (
    TestPingRenderer,
    TestHttpRenderer,
    TestSSLConsistency,
    TestAggregatePing,
    TestRawRenderer,
)


__all__ = [
    TestAggregators,
    TestProbesCommand,
    TestMeasureCommand,
    TestMeasurementsCommand,
    TestReportCommand,
    TestCommandLoading,
    TestArgumentTypeHelper,
    TestPingRenderer,
    TestHttpRenderer,
    TestSSLConsistency,
    TestAggregatePing,
    TestRawRenderer,
]



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

import re

from collections import OrderedDict

from ripe.atlas.cousteau import (
    Ping, Traceroute, Dns, Sslcert, Http, Ntp, AtlasSource, AtlasCreateRequest)

from ...exceptions import RipeAtlasToolsException
from ...helpers.colours import colourise
from ...helpers.validators import ArgumentType
from ...renderers import Renderer
from ...settings import conf, aliases, AliasesDB
from ...streaming import Stream, CaptureLimitExceeded
from ..base import Command as BaseCommand


class Command(BaseCommand):

    NAME = "measure"

    DESCRIPTION = "Create a measurement and optionally wait for the results"

    CREATION_CLASSES = OrderedDict((
        ("ping", Ping),
        ("traceroute", Traceroute),
        ("dns", Dns),
        ("sslcert", Sslcert),
        ("http", Http),
        ("ntp", Ntp)
    ))

    def __init__(self, *args, **kwargs):

        self._type = None
        self._is_oneoff = True

        BaseCommand.__init__(self, *args, **kwargs)

    def _modify_parser_args(self, args):

        kinds = self.CREATION_CLASSES.keys()
        error = (
            "Usage: ripe-atlas measure <{}> [options]\n"
            "\n"
            "  Example: ripe-atlas measure ping --target example.com"
            "".format("|".join(kinds))
        )

        if not args:
            raise RipeAtlasToolsException(error)

        if args[0] not in self.CREATION_CLASSES.keys():
            raise RipeAtlasToolsException(error)
        self._type = args.pop(0)

        if not args:
            args.append("--help")

        return BaseCommand._modify_parser_args(self, args)

    def add_arguments(self):

        self.parser.add_argument(
            "--renderer",
            choices=Renderer.get_available(),
            help="The renderer you want to use. If this isn't defined, an "
                 "appropriate renderer will be selected."
        )
        self.parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Do not create the measurement, only show its definition."
        )

        # Standard for all types

        self.parser.add_argument(
            "--auth",
            type=str,
            default=conf["authorisation"]["create"],
            help="The API key you want to use to create the measurement"
        )
        self.parser.add_argument(
            "--af",
            type=int,
            choices=(4, 6),
            help="The address family, either 4 or 6"
        )
        self.parser.add_argument(
            "--description",
            type=str,
            default=conf["specification"]["description"],
            help="A free-form description"
        )
        self.parser.add_argument(  # Most types
            "--target",
            type=ArgumentType.ip_or_domain,
            help="The target, either a domain name or IP address.  If creating "
                 "a DNS measurement, the absence of this option will imply "
                 "that you wish to use the probe's resolver."
        )
        self.parser.add_argument(
            "--no-report",
            action="store_true",
            help="Don't wait for a response from the measurement, just return "
                 "the URL at which you can later get information about the "
                 "measurement."
        )
        self.parser.add_argument(
            "--set-alias",
            help="After creating the measurement, register an alias for it.",
            type=ArgumentType.alias_is_valid,
            metavar="ALIAS"
        )

        self.parser.add_argument(
            "--interval",
            type=int,
            help="Rather than run this measurement as a one-off (the default), "
                 "create this measurement as a recurring one, with an interval "
                 "of n seconds between attempted measurements. This option "
                 "implies --no-report."
        )

        origins = self.parser.add_mutually_exclusive_group()
        origins.add_argument(
            "--from-area",
            type=str,
            choices=("WW", "West", "North-Central", "South-Central",
                     "North-East", "South-East"),
            help="The area from which you'd like to select your probes."
        )
        origins.add_argument(
            "--from-country",
            type=ArgumentType.country_code,
            metavar="COUNTRY",
            help="The two-letter ISO code for the country from which you'd "
                 "like to select your probes. Example: --from-country=GR"
        )
        origins.add_argument(
            "--from-prefix",
            type=str,
            metavar="PREFIX",
            help="The prefix from which you'd like to select your probes. "
                 "Example: --from-prefix=82.92.0.0/14"
        )
        origins.add_argument(
            "--from-asn",
            # http://www.iana.org/assignments/as-numbers/as-numbers.xhtml
            type=ArgumentType.integer_range(1, 2 ** 32 - 2),
            metavar="ASN",
            help="The ASN from which you'd like to select your probes. "
                 "Example: --from-asn=3333"
        )
        origins.add_argument(
            "--from-probes",
            type=ArgumentType.comma_separated_integers(minimum=1),
            metavar="PROBES",
            help="A comma-separated list of probe-ids you want to use in your "
                 "measurement. Example: --from-probes=1,2,34,157,10006"
        )
        origins.add_argument(
            "--from-measurement",
            type=ArgumentType.integer_range(minimum=1),
            metavar="MEASUREMENT_ID",
            help="A measurement id which you want to use as the basis for "
                 "probe selection in your new measurement.  This is a handy "
                 "way to re-create a measurement under conditions similar to "
                 "another measurement. Example: --from-measurement=1000192"
        )
        self.parser.add_argument(
            "--probes",
            type=ArgumentType.integer_range(minimum=1),
            default=None,
            help="The number of probes you want to use.  Defaults to {},"
                 "unless --from-probes is invoked, in which case the number of "
                 "probes selected is used.".format(conf["specification"]["source"]["requested"])
        )
        self.parser.add_argument(
            "--include-tag",
            type=ArgumentType.tag,
            action="append",
            metavar="TAG",
            help="Include only probes that are marked with these tags. "
                 "Example: --include-tag=system-ipv6-works"
        )
        self.parser.add_argument(
            "--exclude-tag",
            type=ArgumentType.tag,
            action="append",
            metavar="TAG",
            help="Exclude probes that are marked with these tags. "
                 "Example: --exclude-tag=system-ipv6-works"
        )

        self.parser.add_argument(
            "--group-id",
            type=int,
            help="Add newly created measurement to a group that you own",
        )

        # Validation is too complex because it's based on the interval, so
        # we just accept the round-trip for server-side validation
        self.parser.add_argument(
            "--spread",
            type=int,
            default=conf["specification"]["spread"],
            help="Specify the spread of probes within a single measurement interval",
        )
        self.parser.add_argument(
            "--resolve-on-probe",
            action="store_true",
            default=conf["specification"]["resolve_on_probe"],
            help="Causes the target to be resolved by each probe rather than once by the server",
        )

        Renderer.add_arguments_for_available_renderers(self.parser)

    def run(self):

        self._account_for_selected_probes()

        if self.arguments.dry_run:
            return self.dry_run()

        is_success, response = self.create()

        if not is_success:
            self._handle_api_error(response)  # Raises an exception

        pk = response["measurements"][0]
        url = "{0}/measurements/{1}/".format(conf["ripe-ncc"]["endpoint"], pk)

        self.ok(
            "Looking good!  Your measurement was created and details about "
            "it can be found here:\n\n  {0}".format(url)
        )

        if self.arguments.set_alias:
            alias = self.arguments.set_alias
            aliases["measurement"][alias] = pk
            AliasesDB.write(aliases)

        if not self.arguments.no_report:
            self.stream(pk, url)

    def dry_run(self):

        print(colourise("\nDefinitions:\n{}".format("=" * 80), "bold"))

        for param, val in self._get_measurement_kwargs().items():
            print(colourise("{:<25} {}".format(param, val), "cyan"))

        print(colourise("\nSources:\n{}".format("=" * 80), "bold"))

        for param, val in self._get_source_kwargs().items():
            if param == "tags":
                print(colourise("tags\n  include{}{}\n  exclude{}{}\n".format(
                    " " * 17,
                    ", ".join(val["include"]),
                    " " * 17,
                    ", ".join(val["exclude"])
                ), "cyan"))
                continue
            print(colourise("{:<25} {}".format(param, val), "cyan"))

    def create(self):
        creation_class = self.CREATION_CLASSES[self._type]

        return AtlasCreateRequest(
            server=conf["ripe-ncc"]["endpoint"].replace("https://", ""),
            key=self.arguments.auth,
            user_agent=self.user_agent,
            measurements=[creation_class(**self._get_measurement_kwargs())],
            sources=[AtlasSource(**self._get_source_kwargs())],
            is_oneoff=self._is_oneoff
        ).create()

    def stream(self, pk, url):
        self.ok("Connecting to stream...")
        try:
            Stream(capture_limit=self.arguments.probes, timeout=300).stream(
                self.arguments.renderer, self.arguments, self._type, pk)
        except (KeyboardInterrupt, CaptureLimitExceeded):
            pass  # User said stop, so we fall through to the finally block.
        finally:
            self.ok("Disconnecting from stream\n\nYou can find details "
                    "about this measurement here:\n\n  {0}".format(url))

    def clean_target(self):

        if not self.arguments.target:
            raise RipeAtlasToolsException(
                "You must specify a target for that kind of measurement"
            )

        return self.arguments.target

    def clean_description(self):
        if self.arguments.description:
            return self.arguments.description
        if conf["specification"]["description"]:
            return conf["specification"]["description"]
        return "{} measurement to {}".format(
            self._type.capitalize(), self.arguments.target)

    def _get_measurement_kwargs(self):

        # This is kept apart from the r = {} because dns measurements don't
        # require a target attribute
        target = self.clean_target()

        r = {
            "af": self._get_af(),
            "description": self.clean_description(),
        }

        spec = conf["specification"]  # Shorter names are easier to read
        if self.arguments.interval or spec["times"]["interval"]:
            r["interval"] = self.arguments.interval
            self._is_oneoff = False
            self.arguments.no_report = True
        elif not spec["times"]["one-off"]:
            raise RipeAtlasToolsException(
                "Your configuration file appears to be setup to not create "
                "one-offs, but also offers no interval value.  Without one of "
                "these, a measurement cannot be created."
            )

        if target:
            r["target"] = target

        if self.arguments.group_id:
            r["group_id"] = self.arguments.group_id

        if self.arguments.spread is not None:
            r["spread"] = self.arguments.spread

        if self.arguments.resolve_on_probe is not None:
            r["resolve_on_probe"] = self.arguments.resolve_on_probe

        return r

    def _get_source_kwargs(self):

        r = conf["specification"]["source"]

        r["requested"] = self.arguments.probes
        if self.arguments.from_country:
            r["type"] = "country"
            r["value"] = self.arguments.from_country
        elif self.arguments.from_area:
            r["type"] = "area"
            r["value"] = self.arguments.from_area
        elif self.arguments.from_prefix:
            r["type"] = "prefix"
            r["value"] = self.arguments.from_prefix
        elif self.arguments.from_asn:
            r["type"] = "asn"
            r["value"] = self.arguments.from_asn
        elif self.arguments.from_probes:
            r["type"] = "probes"
            r["value"] = ",".join([str(_) for _ in self.arguments.from_probes])
        elif self.arguments.from_measurement:
            r["type"] = "msm"
            r["value"] = self.arguments.from_measurement

        r["tags"] = {
            "include": self.arguments.include_tag or [],
            "exclude": self.arguments.exclude_tag or []
        }

        af = "ipv{}".format(self._get_af())
        kind = self._type
        spec = conf["specification"]
        for clude in ("in", "ex"):
            clude += "clude"
            if not r["tags"][clude]:
                r["tags"][clude] += spec["tags"][af][kind][clude]
                r["tags"][clude] += spec["tags"][af]["all"][clude]

        return r

    def _get_af(self):
        """
        Returns the specified af, or a guessed one, or the configured one.  In
        that order.
        """
        if self.arguments.af:
            return self.arguments.af
        if self.arguments.target:
            if ":" in self.arguments.target:
                return 6
            if re.match(r"^\d+\.\d+\.\d+\.\d+$", self.arguments.target):
                return 4
        return conf["specification"]["af"]

    def _account_for_selected_probes(self):
        """
        If the user has used --from-probes, there's a little extra magic we
        need to do.
        """

        # We can't use argparse's mutually_exclusive_group() method here because
        # that library doesn't allow partial overlap.
        if self.arguments.from_probes and self.arguments.probes:
            raise RipeAtlasToolsException(
                "Explicit probe selection (--from-probes) in incompatible with "
                "a --probes argument."
            )

        configured = conf["specification"]["source"]["requested"]
        if not self.arguments.probes:
            self.arguments.probes = configured
            if self.arguments.from_probes:
                self.arguments.probes = len(self.arguments.from_probes)

    @staticmethod
    def _handle_api_error(response):

        error_detail = response

        if isinstance(response, dict) and "detail" in response:
            error_detail = response["detail"]

        message = (
            "There was a problem communicating with the RIPE Atlas "
            "infrastructure.  The message given was:\n\n  {}"
        ).format(error_detail)

        raise RipeAtlasToolsException(message)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import copy
import os
import re
import yaml


class UserSettingsParser(object):

    USER_CONFIG_DIR = os.path.join(
        os.path.expanduser("~"), ".config", "ripe-atlas-tools")

    USER_RC = None

    def get(self):
        r = copy.deepcopy(self.DEFAULT)
        if os.path.exists(self.USER_RC):
            with open(self.USER_RC) as y:
                custom = yaml.load(y)
                if custom:
                    r = self.deep_update(r, custom)
        return r

    @classmethod
    def deep_update(cls, d, u):
        """
        Updates a dictionary with another dictionary, only it goes deep.
        Stolen from http://stackoverflow.com/questions/3232943/
        """
        for k, v in u.items():
            if isinstance(v, collections.Mapping):
                r = cls.deep_update(d.get(k, {}), v)
                d[k] = r
            else:
                d[k] = u[k]
        return d

    @staticmethod
    def write(data):
        raise NotImplementedError()


class Configuration(UserSettingsParser):
    """
    A singleton configuration class that's smart enough to create a config
    out of defaults + yaml
    """

    USER_RC = os.path.join(UserSettingsParser.USER_CONFIG_DIR, "rc")

    DEFAULT = {
        "authorisation": {
            "fetch": "",
            "fetch_aliases": {},
            "create": "",
        },
        "specification": {
            "af": 4,
            "description": "",
            "source": {
                "type": "area",
                "value": "WW",
                "requested": 50,
            },
            "spread": None,
            "resolve_on_probe": None,
            "times": {
                "one-off": True,
                "interval": None,
                "start": None,
                "stop": None,
            },
            "types": {
                "ping": {
                    "packets": 3,
                    "packet-interval": 1000,
                    "size": 48,
                    "include_probe_id": None,
                },
                "traceroute": {
                    "packets": 3,
                    "size": 48,
                    "protocol": "ICMP",
                    "dont-fragment": False,
                    "paris": 0,
                    "first-hop": 1,
                    "max-hops": 255,
                    "port": 80,
                    "destination-option-size": None,
                    "hop-by-hop-option-size": None,
                    "timeout": 4000,
                    "response-timeout": None,
                    "duplicate-timeout": None,
                },
                "sslcert": {
                    "port": 443,
                    "hostname": None,
                },
                "ntp": {
                    "packets": 3,
                    "timeout": 4000
                },
                "dns": {
                    "set-cd-bit": False,
                    "set-do-bit": False,
                    "protocol": "UDP",
                    "query-class": "IN",
                    "query-type": "A",
                    "query-argument": None,
                    "set-nsid-bit": False,
                    "udp-payload-size": 512,
                    "set-rd-bit": True,
                    "retry": 0,
                    "timeout": None,
                },
                "http": {
                    "header-bytes": 0,
                    "version": "1.1",
                    "method": "GET",
                    "port": 80,
                    "path": "/",
                    "query-string": None,
                    "user-agent": "RIPE Atlas: https://atlas.ripe.net/",
                    "body-bytes": None,
                    "timing-verbosity": 0,
                },
            },
            "tags": {
                "ipv4": {
                    "ping": {
                        "include": [],
                        "exclude": []
                    },
                    "traceroute": {
                        "include": [],
                        "exclude": []
                    },
                    "dns": {
                        "include": [],
                        "exclude": []
                    },
                    "sslcert": {
                        "include": [],
                        "exclude": []
                    },
                    "http": {
                        "include": [],
                        "exclude": []
                    },
                    "ntp": {
                        "include": [],
                        "exclude": []
                    },
                    "all": {
                        "include": ["system-ipv4-works"],
                        "exclude": []
                    },
                },
                "ipv6": {
                    "ping": {
                        "include": [],
                        "exclude": []
                    },
                    "traceroute": {
                        "include": [],
                        "exclude": []
                    },
                    "dns": {
                        "include": [],
                        "exclude": []
                    },
                    "sslcert": {
                        "include": [],
                        "exclude": []
                    },
                    "http": {
                        "include": [],
                        "exclude": []
                    },
                    "ntp": {
                        "include": [],
                        "exclude": []
                    },
                    "all": {
                        "include": ["system-ipv6-works"],
                        "exclude": []
                    }
                }
            }
        },
        "ripe-ncc": {
            "endpoint": "https://atlas.ripe.net",
            "version": 0,
        }
    }

    @staticmethod
    def write(config):
        """
        PyYaml is incapable of preserving comments, or even specifying them as
        an argument to `.dump()` (http://pyyaml.org/ticket/114), so we have to
        do some regex gymnastics here to make sure that the config file remains
        easy for n00bs to read.
        """

        template = os.path.join(
            os.path.dirname(__file__), "templates", "base.yaml")

        authorisation = re.compile("^authorisation:$", re.MULTILINE)
        tags = re.compile("^  tags:$", re.MULTILINE)
        specification = re.compile("^specification:$", re.MULTILINE)
        ripe = re.compile("^ripe-ncc:$", re.MULTILINE)

        with open(template) as t:
            payload = str(t.read()).format(
                payload=yaml.dump(
                    config,
                    default_flow_style=False
                )
            )
            payload = ripe.sub(
                "\n# Don't mess with these, or Bad Things may happen\n"
                "ripe-ncc:",
                payload
            )
            payload = authorisation.sub(
                "# Authorisation\n"
                "authorisation:",
                payload
            )
            payload = specification.sub(
                "\n# Measurement Creation\n"
                "specification:",
                payload
            )
            payload = tags.sub(
                "  # Tags added to probes selection\n"
                "  tags:",
                payload
            )

        with open(Configuration.USER_RC, "w") as rc:
            rc.write(payload)


class AliasesDB(UserSettingsParser):
    """
    A singleton class to manage user aliases
    """

    USER_RC = os.path.join(UserSettingsParser.USER_CONFIG_DIR, "aliases")

    DEFAULT = {
        "measurement": {},
        "probe": {}
    }

    @staticmethod
    def write(aliases):
        if not os.path.exists(AliasesDB.USER_CONFIG_DIR):
            os.makedirs(AliasesDB.USER_CONFIG_DIR)

        payload = yaml.dump(
            aliases,
            default_flow_style=False
        )

        with open(AliasesDB.USER_RC, "w") as rc:
            rc.write(payload)


conf = Configuration().get()
aliases = AliasesDB().get()



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import json

from .base import Renderer as BaseRenderer
from ..helpers.sanitisers import sanitise


class Renderer(BaseRenderer):

    RENDERS = [
        BaseRenderer.TYPE_PING,
        BaseRenderer.TYPE_TRACEROUTE,
        BaseRenderer.TYPE_DNS,
        BaseRenderer.TYPE_SSLCERT,
        BaseRenderer.TYPE_HTTP,
        BaseRenderer.TYPE_NTP
    ]

    SHOW_DEFAULT_HEADER = False
    SHOW_DEFAULT_FOOTER = False

    def on_result(self, result, probes=None):
        return sanitise(
            json.dumps(result.raw_data, separators=(",", ":"))) + "\n"



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import sys
try:
    import ujson as json
except ImportError:
    import json
import itertools

from ripe.atlas.sagan import Result
from ripe.atlas.cousteau import (
    AtlasLatestRequest, AtlasResultsRequest
)

from ..aggregators import RangeKeyAggregator, ValueKeyAggregator, aggregate
from ..exceptions import RipeAtlasToolsException
from ..helpers.rendering import SaganSet, Rendering
from ..helpers.validators import ArgumentType
from ..renderers import Renderer
from .base import Command as BaseCommand
from ..filters import FilterFactory, filter_results
from ..settings import conf


class Command(BaseCommand):

    NAME = "report"

    DESCRIPTION = (
        "Report the results of an existing measurement from the API, "
        "a file or standard input"
    )
    EXTRA_DESCRIPTION = (
        "Examples:\n"
        "  ripe-atlas report 1001 --probes 157,10006\n"
        "  ripe-atlas report --from-file results.json\n"
        "  cat results.json | ripe-atlas report --aggregate-by prefix_v4\n"
    )

    AGGREGATORS = {
        "country": ["probe.country_code", ValueKeyAggregator],
        "rtt-median": [
            "rtt_median",
            RangeKeyAggregator,
            [10, 20, 30, 40, 50, 100, 200, 300]
        ],
        "status": ["probe.status", ValueKeyAggregator],
        "asn_v4": ["probe.asn_v4", ValueKeyAggregator],
        "asn_v6": ["probe.asn_v6", ValueKeyAggregator],
        "prefix_v4": ["probe.prefix_v4", ValueKeyAggregator],
        "prefix_v6": ["probe.prefix_v6", ValueKeyAggregator],
    }

    def add_arguments(self):
        self.parser.add_argument(
            "measurement_id",
            type=ArgumentType.msm_id_or_name(),
            help="The measurement ID or alias to fetch from the results API. "
            "(Conflicts with the --from-file option)",
            nargs="?",
        )
        self.parser.add_argument(
            "--auth",
            type=str,
            choices=conf["authorisation"]["fetch_aliases"].keys(),
            help="The API key alias you want to use to fetch the measurement. "
                 "To configure an API key alias, use "
                 "ripe-atlas configure --set authorisation.fetch_aliases."
                 "ALIAS_NAME=YOUR_KEY"
        )
        self.parser.add_argument(
            "--probes",
            type=ArgumentType.comma_separated_integers_or_file,
            help="Either a comma-separated list of probe ids you want to see "
                 "exclusively, a path to a file containing probe ids (one on "
                 "each line), or \"-\" for standard input in the same format."
        )
        self.parser.add_argument(
            "--renderer",
            choices=Renderer.get_available(),
            help="The renderer you want to use. If this isn't defined, an "
                 "appropriate renderer will be selected."
        )
        self.parser.add_argument(
            "--aggregate-by",
            type=str,
            choices=self.AGGREGATORS.keys(),
            action="append",
            help="Tell the rendering engine to aggregate the results by the "
                 "selected option. Note that if you opt for aggregation, no "
                 "output will be generated until all results are received."
        )
        self.parser.add_argument(
            "--probe-asns",
            type=ArgumentType.comma_separated_integers(
                minimum=1,
                # http://www.iana.org/assignments/as-numbers/as-numbers.xhtml
                maximum=2 ** 32 - 2
            ),
            help="A comma-separated list of probe ASNs you want to see "
                 "exclusively."
        )
        self.parser.add_argument(
            "--start-time",
            type=ArgumentType.datetime,
            help="The start time of the report."
        )
        self.parser.add_argument(
            "--stop-time",
            type=ArgumentType.datetime,
            help="The stop time of the report."
        )

        self.parser.add_argument(
            "--from-file",
            type=ArgumentType.path,
            help='The source of the data to be rendered. '
            '(Conflicts with specifying measurement_id)',
        )

        Renderer.add_arguments_for_available_renderers(self.parser)

    def _get_request_auth(self):
        if self.arguments.auth:
            return conf["authorisation"]["fetch_aliases"][self.arguments.auth]
        else:
            return conf["authorisation"]["fetch"]

    def _get_request(self):

        kwargs = {
            "msm_id": self.arguments.measurement_id,
            "user_agent": self.user_agent
        }
        kwargs["key"] = self._get_request_auth()
        if self.arguments.probes:
            kwargs["probe_ids"] = self.arguments.probes
        if self.arguments.start_time:
            kwargs["start"] = self.arguments.start_time
        if self.arguments.stop_time:
            kwargs["stop"] = self.arguments.stop_time

        if "start" in kwargs or "stop" in kwargs:
            return AtlasResultsRequest(**kwargs)
        return AtlasLatestRequest(**kwargs)

    def run(self):
        if self.arguments.measurement_id and self.arguments.from_file:
            raise RipeAtlasToolsException(
                "You can only specify one of --from-file or "
                "measurement_id, not both."
            )
        if self.arguments.measurement_id:
            results, sample = self._get_results_from_api(
                self.arguments.measurement_id
            )
            use_regular_file = False
        else:
            if self.arguments.from_file:
                use_regular_file = self.arguments.from_file != "-"
            elif sys.stdin.isatty():
                self.parser.print_help()
                sys.exit(1)
            else:
                use_regular_file = False

            results, sample = self._get_results_from_file(
                use_regular_file
            )

        # Sagan calls measurements "ssl" when they are actually "sslcert"
        # so we use .raw_data once we have verified and parsed the sample.
        measurement_type = Result.get(sample).raw_data["type"].lower()

        renderer = Renderer.get_renderer(
            self.arguments.renderer, measurement_type
        )(arguments=self.arguments)

        results = SaganSet(iterable=results, probes=self.arguments.probes)

        if self.arguments.probe_asns:
            asn_filters = set([])
            for asn in self.arguments.probe_asns:
                asn_filters.add(FilterFactory.create("asn", asn))
            results = filter_results(asn_filters, list(results))

        if self.arguments.aggregate_by:
            results = aggregate(results, self.get_aggregators())

        Rendering(renderer=renderer, payload=results).render()

        if use_regular_file:
            self.file.close()

    def _get_results_from_api(self, measurement_id):

        results = self._get_request().get()[1]
        if isinstance(results, list):
            if not results:
                raise RipeAtlasToolsException(
                    "There aren't any results for your request.")
        else:
            error = results.get("error")
            msg = "Error fetching measurement results"
            if error:
                msg += ": [{status} {title}] {detail}".format(
                    **error
                )
            else:
                msg = "{} Error fetching measurement results".format(error)
            raise RipeAtlasToolsException(msg)
        sample = results[0]
        return results, sample

    def _get_results_from_file(self, using_regular_file):
        """
        We need to get the first result from the source in order to detect the
        type.  Additionally, if the source is actually one great big JSON list,
        then we need to parse it so we iterate over the results since there's no
        newline characters.
        """

        self.file = sys.stdin
        if using_regular_file:
            self.file = open(self.arguments.from_file)

        # Pop the first line off the source stack.  This may very well be a Very
        # Large String and cause a memory explosion, but we like to let our
        # users shoot themselves in the foot.
        sample = next(self.file)

        # Re-attach the line back onto the iterable so we don't lose anything
        results = itertools.chain([sample], self.file)

        # In the case of the Very Large String, we parse out the JSON here
        if sample.startswith("["):
            results = json.loads("".join(results))
            sample = results[0]  # Reassign sample to an actual result

        return results, sample

    def get_aggregators(self):
        """Return aggregators list based on user input"""
        aggregation_keys = []
        for aggr_key in self.arguments.aggregate_by:
            # Get class and aggregator key
            aggregation_class = self.AGGREGATORS[aggr_key][1]
            key = self.AGGREGATORS[aggr_key][0]
            if aggr_key == "rtt-median":
                # Get range for the aggregation
                key_range = self.AGGREGATORS[aggr_key][2]
                aggregation_keys.append(
                    aggregation_class(key=key, ranges=key_range)
                )
            else:
                aggregation_keys.append(aggregation_class(key=key))
        return aggregation_keys



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import os
import pkgutil
import sys

from ..exceptions import RipeAtlasToolsException


class Renderer(object):

    TYPE_PING = "ping"
    TYPE_TRACEROUTE = "traceroute"
    TYPE_DNS = "dns"
    TYPE_SSLCERT = "sslcert"
    TYPE_HTTP = "http"
    TYPE_NTP = "ntp"

    RENDERS = ()

    SHOW_DEFAULT_HEADER = True
    SHOW_DEFAULT_FOOTER = True

    def __init__(self, *args, **kwargs):
        """
        If "arguments" is in kwargs it can be used to gather renderer's
        optional arguments which have been passed via CLI.
        See also add_arguments().
        """
        pass

    @staticmethod
    def get_available():
        """
        Return a list of renderers available to be used.
        """

        paths = [os.path.dirname(__file__)]
        if "HOME" in os.environ:
            path = os.path.join(
                os.environ["HOME"], ".config", "ripe-atlas-tools")
            sys.path.append(path)
            paths += [os.path.join(path, "renderers")]

        r = [package_name for _, package_name, _ in pkgutil.iter_modules(paths)]
        r.remove("base")

        return r

    @staticmethod
    def add_arguments_for_available_renderers(parser):
        for renderer_name in Renderer.get_available():
            renderer_cls = Renderer.get_renderer_by_name(renderer_name)
            renderer_cls.add_arguments(parser)

    @staticmethod
    def render(template, **kwargs):
        """
        A crude templating engine.
        """

        template = os.path.join(
            os.path.dirname(__file__), "templates", template)

        with open(template) as f:
            return str(f.read()).format(**kwargs)

    @classmethod
    def get_renderer(cls, name=None, kind=None):
        """
        Using the name if you've asked for it specifically, or attempting to
        guess the appropriate renderer based on the kind of measurement, this
        will return a Renderer subclass or None if nothing can be found.
        """

        renderer = None

        if name:
            renderer = cls.get_renderer_by_name(name)

        if not renderer and kind:
            renderer = cls.get_renderer_by_kind(kind)

        if kind:
            cls._test_renderer_accepts_kind(renderer, kind)

        return renderer

    @classmethod
    def get_renderer_by_name(cls, name):
        error_message = (
            'The renderer you selected, "{}" could not be found.'
        ).format(name)

        try:  # User-defined, user-supplied
            r = cls.import_renderer("renderers", name)
        except ImportError:
            try:  # User-defined, officially-supported
                r = cls.import_renderer("ripe.atlas.tools.renderers", name)
            except ImportError:
                raise RipeAtlasToolsException(error_message)

        return r

    @classmethod
    def get_renderer_by_kind(cls, kind):
        error_message = (
            'The selected renderer, "{}" could not be found.'
        ).format(kind)

        try:
            r = cls.import_renderer("ripe.atlas.tools.renderers", kind)
        except ImportError:
            raise RipeAtlasToolsException(error_message)

        return r

    @staticmethod
    def import_renderer(package, name):
        return getattr(
            importlib.import_module("{}.{}".format(package, name)),
            "Renderer"
        )

    @staticmethod
    def add_arguments(parser):
        """
        Add renderer's optional arguments here.

        Suggested format:

        group = parser.add_argument_group(
            title="Optional arguments for XXX renderer"
        )
        group.add_argument(
            ...
        )
        """
        pass

    def header(self):
        """
        Override this to add an additional header.
        """
        return ""

    def additional(self, results):
        """
        Override this for summary logic.
        """
        return ""

    def footer(self):
        """
        Override this to add an additional footer.
        """
        return ""

    @staticmethod
    def _test_renderer_accepts_kind(renderer, kind):
        if kind not in renderer.RENDERS:
            raise RipeAtlasToolsException(
                'The renderer selected does not appear to support measurements '
                'of type "{}"'.format(kind)
            )

    def on_result(self, result):
        """
        This must be defined in the subclass, and must return a string, even if
        that string is "".
        """
        raise NotImplementedError()


class Result(str):
    """
    A string-like object that we can use to render results, but that contains
    enough information to be used by the aggregators if need be.
    """
    def __new__(cls, value, probe_id):
        obj = str.__new__(cls, value)
        obj.probe_id = probe_id
        return obj



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from ripe.atlas.sagan import Result
from ripe.atlas.tools.renderers.ping import Renderer


class TestPingRenderer(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.basic = Result.get('{"af":4,"prb_id":1,"result":[{"rtt":10.001},{"rtt":10.002},{"rtt":10.003}],"ttl":20,"avg":10.002,"size":20,"from":"1.2.3.4","proto":"ICMP","timestamp":1440000000,"dup":0,"type":"ping","sent":3,"msm_id":1000001,"fw":4700,"max":10.003,"step":360,"src_addr":"2.3.4.5","rcvd":3,"msm_name":"Ping","lts":40,"dst_name":"my.name.ca","min":10.001,"dst_addr":"3.4.5.6"}')
        self.no_packets = Result.get('{"af":4,"prb_id":2,"result":[],"ttl":20,"avg":10.002,"size":20,"from":"1.2.3.4","proto":"ICMP","timestamp":1440000000,"dup":0,"type":"ping","sent":3,"msm_id":1000001,"fw":4700,"max":null,"step":360,"src_addr":"2.3.4.5","rcvd":0,"msm_name":"Ping","lts":40,"dst_name":"my.name.ca","min":null,"dst_addr":"3.4.5.6"}')

    def test_basic(self):
        self.assertEqual(
            Renderer().on_result(self.basic),
            "20 bytes from probe #1     1.2.3.4         to my.name.ca (3.4.5.6): ttl=20 times:10.001,  10.002,  10.003, \n"
        )

    def test_no_packets(self):
        self.assertEqual(
            Renderer().on_result(self.no_packets), "No packets found\n")



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from ripe.atlas.sagan import Result
from ripe.atlas.cousteau import Probe
from ripe.atlas.tools.exceptions import RipeAtlasToolsException
from ripe.atlas.tools.filters import (
    FilterFactory, Filter, ASNFilter, filter_results
)


class TestFilterFactory(unittest.TestCase):
    def test_factory_create(self):
        """Test factory create."""
        self.assertIsInstance(FilterFactory.create("country", "GR"), Filter)
        self.assertIsInstance(FilterFactory.create("asn", "3333"), ASNFilter)


class TestFilter(unittest.TestCase):

    def test_filter(self):
        """Tests filter method of general Filter class."""
        result = {'af': 4, 'prb_id': 1216, 'result': [{'rtt': 27.429}, {'rtt': 25.672}, {'rtt': 25.681}], 'ttl': 54, 'avg': 26.2606666667, 'size': 20, 'from': '109.190.83.40', ' proto': 'ICMP', 'timestamp': 1445025400, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4700, 'max': 27.429, 'step': 360, 'src_addr': '192.168.103.132', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 377, 'dst_name': 'hsi.cablecom.ch', 'min': 25.672, 'dst_addr': '62.2.16.24'}
        sagan_result = Result.get(
            result, on_error=Result.ACTION_IGNORE, on_warning=Result.ACTION_IGNORE
        )
        sagan_result.probe = Probe(
            id=1216,
            meta_data={"country_code": "GR", "asn_v4": 3337, "asn_v6": "4445"}
        )
        self.assertTrue(Filter("country_code", "GR").filter(sagan_result))
        self.assertFalse(Filter("country_code", "NL").filter(sagan_result))
        self.assertFalse(Filter("asn_v4", 3336).filter(sagan_result))
        self.assertTrue(Filter("asn_v6", "4445").filter(sagan_result))

        with self.assertRaises(RipeAtlasToolsException):
            self.assertTrue(Filter("country", "GR").filter(sagan_result))


class TestASNFilter(unittest.TestCase):

    def test_filter(self):
        """Tests filter method of probe's asn filter class."""
        result = {'af': 4, 'prb_id': 1216, 'result': [{'rtt': 27.429}, {'rtt': 25.672}, {'rtt': 25.681}], 'ttl': 54, 'avg': 26.2606666667, 'size': 20, 'from': '109.190.83.40', ' proto': 'ICMP', 'timestamp': 1445025400, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4700, 'max': 27.429, 'step': 360, 'src_addr': '192.168.103.132', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 377, 'dst_name': 'hsi.cablecom.ch', 'min': 25.672, 'dst_addr': '62.2.16.24'}
        sagan_result = Result.get(
            result, on_error=Result.ACTION_IGNORE, on_warning=Result.ACTION_IGNORE
        )
        sagan_result.probe = Probe(
            id=1216,
            meta_data={"country_code": "GR", "asn_v4": 3337, "asn_v6": "4445"}
        )
        self.assertTrue(ASNFilter(3337).filter(sagan_result))
        self.assertFalse(ASNFilter(3336).filter(sagan_result))
        self.assertTrue(ASNFilter("4445").filter(sagan_result))
        self.assertFalse(ASNFilter(4445).filter(sagan_result))


class TestFilterResults(unittest.TestCase):
    def setUp(self):
        result = {'af': 4, 'prb_id': 1216, 'result': [{'rtt': 27.429}, {'rtt': 25.672}, {'rtt': 25.681}], 'ttl': 54, 'avg': 26.2606666667, 'size': 20, 'from': '109.190.83.40', ' proto': 'ICMP', 'timestamp': 1445025400, 'dup': 0, 'type': 'ping', 'sent': 3, 'msm_id': 1000192, 'fw': 4700, 'max': 27.429, 'step': 360, 'src_addr': '192.168.103.132', 'rcvd': 3, 'msm_name': 'Ping', 'lts': 377, 'dst_name': 'hsi.cablecom.ch', 'min': 25.672, 'dst_addr': '62.2.16.24'}

        probes = [
            Probe(
                id=1216,
                meta_data={"country_code": "GR", "asn_v4": 3337, "asn_v6": "4445"}
            ),
            Probe(
                id=121,
                meta_data={"country_code": "GR", "asn_v4": 3338, "asn_v6": "4445"}
            ),
            Probe(
                id=12,
                meta_data={"country_code": "DE", "asn_v4": 3339, "asn_v6": 3337}
            ),
            Probe(
                id=1,
                meta_data={"country_code": "NL", "asn_v4": 3337, "asn_v6": "4446"}
            )
        ]
        self.sagan_results = []
        for probe in probes:
            sagan_result = Result.get(
                result, on_error=Result.ACTION_IGNORE, on_warning=Result.ACTION_IGNORE
            )
            sagan_result.probe = probe
            self.sagan_results.append(sagan_result)

    def test_filter_results1(self):
        """Tests filter results where we have 3 results match with 2 filters."""
        expected_results = [
            self.sagan_results[0], self.sagan_results[2], self.sagan_results[3]
        ]
        filters = [
            FilterFactory.create("asn", 3337),
            FilterFactory.create("country_code", "NL")
        ]
        self.assertEqual(
            filter_results(filters, self.sagan_results), expected_results
        )

    def test_filter_results2(self):
        """Tests filter results where we have no results match with multiple filters."""
        expected_results = []
        filters = [
            FilterFactory.create("asn", 3336),
            FilterFactory.create("country_code", "DK"),
            FilterFactory.create("asn_v6", "3")
        ]
        self.assertEqual(
            filter_results(filters, self.sagan_results), expected_results
        )

    def test_filter_results3(self):
        """Tests filter results where we have success from one of the filters."""
        expected_results = [self.sagan_results[3]]
        filters = [
            FilterFactory.create("asn", 3336),
            FilterFactory.create("country_code", "NL")
        ]
        self.assertEqual(
            filter_results(filters, self.sagan_results), expected_results
        )

    def test_filter_results4(self):
        """Tests filter results where we have success from both of the filters."""
        expected_results = [self.sagan_results[3]]
        filters = [
            FilterFactory.create("asn_v6", "4446"),
            FilterFactory.create("country_code", "NL")
        ]
        self.assertEqual(
            filter_results(filters, self.sagan_results), expected_results
        )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..helpers.colours import colourise

from .base import Renderer as BaseRenderer


class Renderer(BaseRenderer):
    """
    We're abusing the Extended Log File Format here to render the result,
    amending it to include a few things not originally specified in the W3C
    Working Draft: http://www.w3.org/TR/WD-logfile.html Namely:
      http-version, header-bytes, and body-bytes
    """

    RENDERS = [BaseRenderer.TYPE_HTTP]
    COLOURS = {
        "2": "green",
        "3": "blue",
        "4": "yellow",
        "5": "red"
    }

    def on_result(self, result, probes=None):
        r = "#Version: 1.0\n#Date: {}\n#Fields: {}\n".format(
            result.created.strftime("%Y-%m-%d %H:%M:%S"),
            "cs-method cs-uri c-ip s-ip sc-status time-taken http-version "
            "header-bytes body-bytes"
        )
        for response in result.responses:
            r += self._colourise_by_status(
                "{} {} {} {} {} {} {} {} {}\n".format(
                    result.method,
                    result.uri,
                    response.source_address,
                    response.destination_address,
                    response.code,
                    response.response_time,
                    response.version,
                    response.head_size,
                    response.body_size
                ),
                response.code
            )

        return r + "\n"

    def _colourise_by_status(self, output, status):
        try:
            return colourise(output, self.COLOURS[str(status)[0]])
        except (IndexError, KeyError):
            return colourise(output, "red")



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ripe.atlas.cousteau import Probe
from ripe.atlas.cousteau.exceptions import APIResponseError

from .base import Command as BaseCommand, MetaDataMixin
from ..exceptions import RipeAtlasToolsException
from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType


class Command(MetaDataMixin, BaseCommand):

    NAME = "probe-info"
    DESCRIPTION = "Return the meta data for one probe"

    def add_arguments(self):
        self.parser.add_argument("id", type=ArgumentType.probe_id_or_name(),
                                 help="The probe id or alias")

    def run(self):

        try:
            probe = Probe(id=self.arguments.id, user_agent=self.user_agent)
        except APIResponseError:
            raise RipeAtlasToolsException(
                "That probe does not appear to exist")

        url_template = "https://atlas.ripe.net/probes/{}/"
        keys = (
            ("id", "ID"),
            ("id", "URL", lambda _: colourise(url_template.format(_), "cyan")),
            ("is_public", "Public?", self._prettify_boolean),
            ("is_anchor", "Anchor?", self._prettify_boolean),
            ("country_code", "Country"),
            ("description", "Description", sanitise),
            ("asn_v4", "ASN (IPv4)"),
            ("asn_v6", "ASN (IPv6)"),
            ("address_v4", "Address (IPv4)"),
            ("address_v6", "Address (IPv6)"),
            ("prefix_v4", "Prefix (IPv4)"),
            ("prefix_v6", "Prefix (IPv6)"),
            ("geometry", "Coordinates", self._prettify_coordinates),
            ("status", "Status"),
        )
        for key in keys:

            value = getattr(probe, key[0])

            if value is None:
                value = "-"
            elif len(key) == 3:
                value = key[2](value)

            self._render_line(key[1], value)

        print(colourise("Tags", "bold"))
        for tag in probe.tags:
            log_tag = "  {}".format(tag["slug"])
            print(log_tag.encode("utf8"))

    @staticmethod
    def _prettify_coordinates(geometry):
        if geometry and "coordinates" in geometry and geometry["coordinates"]:
            return "{},{}".format(
                geometry["coordinates"][1],
                geometry["coordinates"][0]
            )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import sys

from ripe.atlas.cousteau import AtlasStream
from ripe.atlas.sagan import Result

from .renderers import Renderer


class CaptureLimitExceeded(Exception):
    pass


class Stream(object):

    def __init__(self, capture_limit=None, timeout=None):

        self.captured = 0
        self.capture_limit = capture_limit

        self.timeout = timeout

    def stream(self, renderer_name, arguments, kind, pk):

        renderer = Renderer.get_renderer(name=renderer_name, kind=kind)(
            arguments=arguments
        )

        def on_result_response(result, *args):
            sys.stdout.write(renderer.on_result(Result.get(
                result,
                on_error=Result.ACTION_IGNORE,
                on_malformation=Result.ACTION_IGNORE
            )))
            self.captured += 1
            if self.capture_limit and self.captured >= self.capture_limit:
                raise CaptureLimitExceeded()

        stream = AtlasStream()
        stream.connect()

        stream.bind_channel("atlas_result", on_result_response)
        try:
            stream.start_stream(stream_type="result", msm=pk)
            stream.timeout(self.timeout)
        except (KeyboardInterrupt, CaptureLimitExceeded) as e:
            stream.disconnect()
            raise e



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ripe.atlas.sagan.dns import Message

from ...exceptions import RipeAtlasToolsException
from ...helpers.validators import ArgumentType
from ...settings import conf

from .base import Command


class DnsMeasureCommand(Command):

    def _upper_str(self, s):
        """
        Private method to validate specific command line arguments that
        should be provided in upper or lower case
        :param s: string
        :return: string in upper case
        """
        return s.upper()

    def add_arguments(self):

        Command.add_arguments(self)

        specific = self.parser.add_argument_group("DNS-specific Options")
        specific.add_argument(
            "--protocol",
            type=self._upper_str,
            choices=("UDP", "TCP"),
            default=conf["specification"]["types"]["dns"]["protocol"],
            help="The protocol used."
        )
        specific.add_argument(
            "--query-class",
            type=self._upper_str,
            choices=("IN", "CHAOS"),
            default=conf["specification"]["types"]["dns"]["query-class"],
            help='The query class.  The default is "{}"'.format(
                conf["specification"]["types"]["dns"]["query-class"]
            )
        )
        specific.add_argument(
            "--query-type",
            type=self._upper_str,
            choices=list(Message.ANSWER_CLASSES.keys()) + ["ANY"],  # The only ones we can parse
            default=conf["specification"]["types"]["dns"]["query-type"],
            help='The query type.  The default is "{}"'.format(
                conf["specification"]["types"]["dns"]["query-type"]
            )
        )
        specific.add_argument(
            "--query-argument",
            type=str,
            default=conf["specification"]["types"]["dns"]["query-argument"],
            help="The DNS label to query"
        )
        specific.add_argument(
            "--set-cd-bit",
            action="store_true",
            default=conf["specification"]["types"]["dns"]["set-cd-bit"],
            help="Set the DNSSEC Checking Disabled flag (RFC4035)"
        )
        specific.add_argument(
            "--set-do-bit",
            action="store_true",
            default=conf["specification"]["types"]["dns"]["set-do-bit"],
            help="Set the DNSSEC OK flag (RFC3225)"
        )
        specific.add_argument(
            "--set-nsid-bit",
            action="store_true",
            default=conf["specification"]["types"]["dns"]["set-nsid-bit"],
            help="Include an EDNS name server ID request with the query"
        )
        specific.add_argument(
            "--set-rd-bit",
            action="store_true",
            default=conf["specification"]["types"]["dns"]["set-rd-bit"],
            help="Set the Recursion Desired flag"
        )
        specific.add_argument(
            "--retry",
            type=ArgumentType.integer_range(minimum=0, maximum=10),
            default=conf["specification"]["types"]["dns"]["retry"],
            help="Number of times to retry"
        )
        specific.add_argument(
            "--udp-payload-size",
            type=ArgumentType.integer_range(minimum=512, maximum=4096),
            default=conf["specification"]["types"]["dns"]["udp-payload-size"],
            help="May be any integer between 512 and 4096 inclusive"
        )
        specific.add_argument(
            "--timeout",
            default=conf["specification"]["types"]["dns"]["timeout"],
            type=ArgumentType.integer_range(minimum=100, maximum=30000),
            help="Per packet timeout in milliseconds",
        )

    def clean_target(self):
        """
        Targets aren't required for this type
        """
        return self.arguments.target

    def clean_description(self):
        if self.arguments.target:
            return Command.clean_description(self)
        return "DNS measurement for {}".format(self.arguments.query_argument)

    def _get_measurement_kwargs(self):

        r = Command._get_measurement_kwargs(self)

        for opt in ("class", "type", "argument"):
            if not getattr(self.arguments, "query_{0}".format(opt)):
                raise RipeAtlasToolsException(
                    "At a minimum, DNS measurements require a query argument.")

        r["query_class"] = self.arguments.query_class
        r["query_type"] = self.arguments.query_type
        r["query_argument"] = self.arguments.query_argument
        r["set_cd_bit"] = self.arguments.set_cd_bit
        r["set_do_bit"] = self.arguments.set_do_bit
        r["set_rd_bit"] = self.arguments.set_rd_bit
        r["set_nsid_bit"] = self.arguments.set_nsid_bit
        r["protocol"] = self.arguments.protocol
        r["retry"] = self.arguments.retry
        r["udp_payload_size"] = self.arguments.udp_payload_size
        r["use_probe_resolver"] = "target" not in r
        if self.arguments.timeout is not None:
            r["timeout"] = self.arguments.timeout

        return r



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from collections import namedtuple
from ripe.atlas.tools.renderers.aggregate_ping import Renderer
from ripe.atlas.sagan import Result

from ..base import capture_sys_output


class TestAggregatePing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.results = [
            {u'af': 4, u'prb_id': 11421, u'result': [{u'rtt': 42.342895}, {u'rtt': 42.220215}, {u'rtt': 42.40614}], u'ttl': 53, u'avg': 42.3230833333, u'size': 48, u'from': u'178.11.85.39', u'proto': u'ICMP', u'timestamp': 1446146495, u'dup': 0, u'type': u'ping', u'sent': 3, u'msm_id': 2882184, u'fw': 4720, u'max': 42.40614, u'step': 240, u'src_addr': u'192.168.2.101', u'rcvd': 3, u'msm_name': u'Ping', u'lts': 3, u'dst_name': u'194.88.241.228', u'min': 42.220215, u'group_id': 2882184, u'dst_addr': u'194.88.241.228'},
            {u'af': 4, u'prb_id': 11779, u'result': [{u'rtt': 76.61127}, {u'rtt': 76.38997}, {u'rtt': 76.47354}], u'ttl': 51, u'avg': 76.4915933333, u'size': 48, u'from': u'79.106.99.242', u'proto': u'ICMP', u'timestamp': 1446146494, u'dup': 0, u'type': u'ping', u'sent': 3, u'msm_id': 2882184, u'fw': 4720, u'max': 76.61127, u'step': 240, u'src_addr': u'192.168.1.110', u'rcvd': 3, u'msm_name': u'Ping', u'lts': 127, u'dst_name': u'194.88.241.228', u'min': 76.38997, u'group_id': 2882184, u'dst_addr': u'194.88.241.228'},
            {u'af': 4, u'prb_id': 17854, u'result': [{u'rtt': 154.118}, {u'rtt': 154.197}, {u'rtt': 154.845}], u'ttl': 55, u'avg': 154.3866666667, u'size': 48, u'from': u'134.202.20.3', u'proto': u'ICMP', u'timestamp': 1446146493, u'dup': 0, u'type': u'ping', u'sent': 3, u'msm_id': 2882184, u'fw': 4700, u'max': 154.845, u'step': 240, u'src_addr': u'134.202.20.3', u'rcvd': 3, u'msm_name': u'Ping', u'lts': 1, u'dst_name': u'194.88.241.228', u'min': 154.118, u'group_id': 2882184, u'dst_addr': u'194.88.241.228'},
            {u'af': 4, u'prb_id': 3183, u'result': [{u'rtt': 42.263816}, {u'rtt': 42.196233}, {u'rtt': 42.342921}], u'ttl': 56, u'avg': 42.2676566667, u'size': 48, u'from': u'212.122.42.107', u'proto': u'ICMP', u'timestamp': 1446146494, u'dup': 0, u'type': u'ping', u'sent': 3, u'msm_id': 2882184, u'fw': 4720, u'max': 42.342921, u'step': 240, u'src_addr': u'192.168.189.1', u'rcvd': 3, u'msm_name': u'Ping', u'lts': 3, u'dst_name': u'194.88.241.228', u'min': 42.196233, u'group_id': 2882184, u'dst_addr': u'194.88.241.228'},
            {u'af': 4, u'prb_id': 3207, u'result': [{u'rtt': 218.077484}, {u'rtt': 36.921608}, {u'rtt': 38.99444}], u'ttl': 49, u'avg': 97.997844, u'size': 48, u'from': u'134.3.245.244', u'proto': u'ICMP', u'timestamp': 1446146494, u'dup': 0, u'type': u'ping', u'sent': 3, u'msm_id': 2882184, u'fw': 4720, u'max': 218.077484, u'step': 240, u'src_addr': u'10.30.0.4', u'rcvd': 3, u'msm_name': u'Ping', u'lts': 70, u'dst_name': u'194.88.241.228', u'min': 36.921608, u'group_id': 2882184, u'dst_addr': u'194.88.241.228'}
        ]
        cls.sagans = [
            Result.get(
                result,
                on_error=Result.ACTION_IGNORE,
                on_warning=Result.ACTION_IGNORE
            ) for result in cls.results
        ]

    def test_additional(self):
        """Tests whole functionality of additional unit."""
        expected_output = (
            "-- 194.88.241.228 ping statistics ---\n"
            "15 packets transmitted, 15 received, 0.0% loss\n"
            "rtt min/med/avg/max = 36.921608/42.406/82.693/218.077484 ms\n\n"
        )

        self.assertEquals(Renderer().additional(self.sagans), expected_output)

    def test_collect_stats(self):
        """Tests collect stats function."""

        renderer = Renderer()
        renderer.collect_stats(self.sagans)
        self.assertEquals(
            renderer.rtts,
            [
                42.343, 42.22, 42.406, 76.611, 76.39, 76.474, 154.118, 154.197,
                154.845, 42.264, 42.196, 42.343, 218.077, 36.922, 38.994
            ]
        )
        self.assertEquals(renderer.target, "194.88.241.228")
        self.assertEquals(renderer.sent_packets, 15)
        self.assertEquals(renderer.received_packets, 15)
        self.assertEquals(
            renderer.rtts_min,
            [42.220215, 76.38997, 154.118, 42.196233, 36.921608]
        )
        self.assertEquals(
            renderer.rtts_max,
            [42.40614, 76.61127, 154.845, 42.342921, 218.077484]
        )

    def test_collect_min_max_rtts(self):
        """Test use cases for collecting min max rtts."""
        renderer = Renderer()
        renderer.collect_min_max_rtts("min", 3)
        self.assertEquals(renderer.rtts_min, [3])
        renderer.collect_min_max_rtts("min", None)
        self.assertEquals(renderer.rtts_min, [3, 0])

        renderer.collect_min_max_rtts("max", 3)
        self.assertEquals(renderer.rtts_max, [3])
        renderer.collect_min_max_rtts("max", None)
        self.assertEquals(renderer.rtts_max, [3, 0])

    def test_collect_packets_rtt(self):
        """Test use cases for collecting rtts."""
        Packet = namedtuple("Packet", "rtt")
        packets = [Packet(rtt=2), Packet(rtt=3.2), Packet(rtt=5.0)]
        renderer = Renderer()
        renderer.collect_packets_rtt(packets)
        self.assertEquals(renderer.rtts, [2, 3.2, 5.0])

        packets = [Packet(rtt=None), Packet(rtt=3.2), Packet(rtt=5.0)]
        renderer = Renderer()
        renderer.collect_packets_rtt(packets)
        self.assertEquals(renderer.rtts, [0, 3.2, 5.0])

    def test_set_target(self):
        """Tests setting the target."""
        Sagan = namedtuple("Sagan", "destination_name")
        renderer = Renderer()
        sagan = Sagan(destination_name="1")
        renderer.set_target(sagan)
        self.assertEquals(renderer.target, "1")
        sagan = Sagan(destination_name="2")
        renderer.set_target(sagan)
        self.assertEquals(renderer.target, "1")

    def test_calculate_loss(self):
        """Test use cases for calculating loss."""
        renderer = Renderer()
        renderer.sent_packets = 10
        renderer.received_packets = 9
        self.assertEquals(renderer.calculate_loss(), 9.999999999999998)
        renderer.sent_packets = 0
        self.assertEquals(renderer.calculate_loss(), 0)
        renderer.received_packets = 0
        renderer.sent_packets = 10
        self.assertEquals(renderer.calculate_loss(), 100)
        renderer.received_packets = 5
        self.assertEquals(renderer.calculate_loss(), 50)

    def test_mean(self):
        """Test use cases for calculating mean."""
        renderer = Renderer()
        renderer.rtts = [0, 2.0, 5.0, 20]
        self.assertEquals(renderer.mean(), 6.75)
        renderer.rtts = [0, 2.0, 7.5, 5.0, 20]
        self.assertEquals(renderer.mean(), 6.9)
        renderer.rtts = [0, 2.0, 7.5, 5.0, 20, 50]
        self.assertEquals(renderer.mean(), 14.083)

    def test_median(self):
        """Test use cases for calculating median."""
        renderer = Renderer()
        renderer.rtts = [0, 2.0, 5.0, 20]
        self.assertEquals(renderer.median(), 3.5)
        renderer.rtts = [0, 2.0, 7.5, 5.0, 20]
        self.assertEquals(renderer.median(), 5)
        renderer.rtts = [0, 2.0, 7.5, 5.0, 20, 50]
        self.assertEquals(renderer.median(), 6.25)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .base import Renderer as BaseRenderer
from collections import Counter

from ..helpers.sanitisers import sanitise
from ..ipdetails import IP


class Renderer(BaseRenderer):

    RENDERS = [BaseRenderer.TYPE_PING]

    SHOW_DEFAULT_HEADER = False
    SHOW_DEFAULT_FOOTER = False

    def __init__(self, *args, **kwargs):

        BaseRenderer.__init__(self, *args, **kwargs)

        # Keys are timestamps, data struct captures ASN membership
        self.asns = Counter()
        self.asn2name = {}

    def on_result(self, result):
        dst = result.destination_address
        if dst is not None:
            ip = IP(dst)
            if ip.asn:
                self.asns[ip.asn] += 1
                self.asn2name[ip.asn] = sanitise(ip.holder)
            else:
                self.asns['<unknown>'] += 1
                self.asn2name['<unknown>'] = 'unknown'
            return ""
        return ""

    def additional(self, results):

        total = sum(self.asns.values())

        r = ""
        for asn, count in self.asns.most_common():
            r += "AS%s %.2f%% (%s)" % (
                asn,
                100.0 * count / total,
                self.asn2name[asn]
            )

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

import itertools
import six
import requests
import sys

from ripe.atlas.cousteau import ProbeRequest
from ripe.atlas.tools.aggregators import ValueKeyAggregator, aggregate

from .base import Command as BaseCommand, TabularFieldsMixin
from ..exceptions import RipeAtlasToolsException
from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType


# Unknown latitude-longitude coordinates.
UNK_COORDS = -1111.0, -1111.0


class Command(TabularFieldsMixin, BaseCommand):

    NAME = "probe-search"

    DESCRIPTION = (
        "Fetch and print probes fulfilling specified criteria based on "
        "given filters"
    )

    # Column name: (alignment, width)
    COLUMNS = {
        "id": ("<", 5),
        "asn_v4": ("<", 6),
        "asn_v6": ("<", 6),
        "country": ("^", 7),
        "status": ("<", 15),
        "prefix_v4": ("<", 18),
        "prefix_v6": ("<", 18),
        "coordinates": ("<", 19),
        "is_public": ("^", 6),
        "description": ("<", 30),
        "address_v4": ("<", 15),
        "address_v6": ("<", 39),
        "is_anchor": ("^", 6),
    }

    def __init__(self, *args, **kwargs):
        BaseCommand.__init__(self, *args, **kwargs)
        self.aggregators = []
        self.first_line_padding = False

    def add_arguments(self):
        """Adds all commands line arguments for this command."""
        asn = self.parser.add_argument_group("ASN")
        asn.add_argument(
            "--asn",
            type=int,
            help="ASN"
        )
        asn.add_argument(
            "--asnv4",
            type=int,
            help="ASNv4"
        )
        asn.add_argument(
            "--asnv6",
            type=int,
            help="ASNv6"
        )

        prefix = self.parser.add_argument_group("Prefix")
        prefix.add_argument(
            "--prefix",
            type=str,
            help="Prefix"
        )
        prefix.add_argument(
            "--prefixv4",
            type=str,
            help="Prefixv4"
        )
        prefix.add_argument(
            "--prefixv6",
            type=str,
            help="Prefixv6"
        )

        area = self.parser.add_argument_group("Area")
        geo_location = area.add_mutually_exclusive_group()
        geo_location.add_argument(
            "--location",
            type=str,
            help="The location of probes as a string i.e. 'Amsterdam'"
        )
        geo_location.add_argument(
            "--center",
            type=str,
            help="location as <lat>,<lon>-string, i.e. '48.45,9.16'"
        )
        geo_location.add_argument(
            "--country",
            type=str,
            help="The country code of probes."
        )
        area.add_argument(
            "--radius",
            type=int,
            default=15,
            help="Radius in km from specified center/point. Default is 15."
        )

        self.parser.add_argument(
            "--tag",
            type=ArgumentType.tag,
            action="append",
            metavar="TAG",
            help="Include only probes that are marked with these tags. "
                 "Use --tag multiple times to filter on the basis of more "
                 "than one tag. "
                 "Example: --tag system-ipv6-works --tag system-ipv4-works",
            dest="tags"
        )

        self.parser.add_argument(
            "--limit",
            type=int,
            default=25,
            help="Return limited number of probes"
        )
        self.parser.add_argument(
            "--field",
            type=str,
            action="append",
            choices=self.COLUMNS.keys(),
            default=[],
            help="The field(s) to display. Invoke multiple times for multiple "
                 "fields. The default is id, asn_v4, asn_v6, country, and "
                 "status."
        )
        self.parser.add_argument(
            "--aggregate-by",
            type=str,
            choices=[
                'country',
                'asn_v4', 'asn_v6',
                'prefix_v4', 'prefix_v6'
            ],
            action="append",
            help=(
                "Aggregate list of probes based on all specified aggregations."
                " Multiple aggregations supported."
            )
        )
        self.parser.add_argument(
            "--all",
            action='store_true',
            help="Fetch *ALL* probes. That will give you a loooong list."
        )
        self.parser.add_argument(
            "--max-per-aggregation",
            type=int,
            help="Maximum number of probes per aggregated bucket."
        )
        self.parser.add_argument(
            "--ids-only",
            action='store_true',
            help=(
                "Print only IDs of probes. Useful to pipe it to another "
                "command."
            )
        )
        self.parser.add_argument(
            "--status",
            type=int,
            choices=[0, 1, 2, 3],
            help=(
                "Probe's connection status [0 - Never Connected, "
                "1 - Connected, 2 - Disconnected, 3 - Abandoned]"
            )
        )

    def run(self):

        if not self.arguments.field:
            self.arguments.field = (
                "id", "asn_v4", "asn_v6", "country", "status")

        if self.arguments.all:
            self.arguments.limit = sys.maxsize if six.PY3 else sys.maxint

        filters = self.build_request_args()

        if not filters and not self.arguments.all:
            raise RipeAtlasToolsException(colourise(
                "Typically you'd want to run this with some arguments to "
                "filter the probe \nlist, as fetching all of the probes can "
                "take a Very Long Time.  However, if you \ndon't care about "
                "the wait, you can use --all and go get yourself a coffee.",
                "blue"
            ))

        self.set_aggregators()
        probes = ProbeRequest(
            return_objects=True, user_agent=self.user_agent, **filters)
        truncated_probes = itertools.islice(probes, self.arguments.limit)

        if self.arguments.ids_only:
            for probe in truncated_probes:
                print(probe.id)
            return

        hr = self._get_horizontal_rule()

        print(self._get_filter_display(filters))
        print(colourise(self._get_header(), "bold"))
        print(colourise(hr, "bold"))

        if self.arguments.aggregate_by:

            buckets = aggregate(list(truncated_probes), self.aggregators)
            self.render_aggregation(buckets)

        else:

            for probe in truncated_probes:
                print(self._get_line(probe).encode("utf8"))

        print(colourise(hr, "bold"))

        # Print total count of found measurements
        print(("{:>" + str(len(hr)) + "}\n").format(
            "Showing {} of {} total probes".format(
                min(self.arguments.limit, probes.total_count) or "all",
                probes.total_count
            )
        ))

    def render_aggregation(self, aggregation_data, indent=0):
        """
        Recursively traverses through aggregation data and print them indented.
        """

        if isinstance(aggregation_data, dict):

            for k, v in aggregation_data.items():

                if not indent:
                    if self.first_line_padding:
                        print("")
                    else:
                        self.first_line_padding = True

                print((u" " * indent) + colourise(k, "bold"))
                self.render_aggregation(v, indent=indent + 1)

        elif isinstance(aggregation_data, list):

            for index, probe in enumerate(aggregation_data):
                print(" {}".format(self._get_line(probe)).encode("utf8"))
                if self.arguments.max_per_aggregation:
                    if index >= self.arguments.max_per_aggregation - 1:
                        break

    def build_request_args(self):
        """
        Builds the request arguments from parser arguments and returns a dict
        that can be used with ATLAS API.
        """
        if self.arguments.all:
            return {}

        return self._clean_request_args()

    def _clean_request_args(self):
        """Cleans all arguments for the API request and checks for sanity."""
        args = {}

        if any(
            [self.arguments.asn, self.arguments.asnv4, self.arguments.asnv6]
        ):
            args.update(self._clean_asn())

        if any([
            self.arguments.prefix,
            self.arguments.prefixv4,
            self.arguments.prefixv6
        ]):
            args.update(self._clean_prefix())

        if self.arguments.location:
            args.update(self._clean_location())

        if self.arguments.center:
            args.update(self._clean_center())

        if self.arguments.country:
            args.update(self._clean_country_code())

        if self.arguments.status is not None:
            args.update({"status": self.arguments.status})

        if self.arguments.tags:
            args.update({"tags": ",".join(self.arguments.tags)})

        return args

    def _clean_asn(self):
        """Make sure ASN arguments don't conflict and make sense."""
        asn = self.arguments.asn
        asnv4 = self.arguments.asnv4
        asnv6 = self.arguments.asnv6

        if asn and (asnv4 or asnv6):
            exc_log = (
                "Specifying argument --asn together with --asnv4/--asnv6 "
                "doesn't make sense"
            )
            raise RipeAtlasToolsException(exc_log)
        if asn:
            return {"asn": asn}

        asn_args = {}
        if asnv4:
            asn_args["asn_v4"] = asnv4

        if asnv6:
            asn_args["asn_v6"] = asnv6

        return asn_args

    def _clean_prefix(self):
        """Make sure ASN arguments don't conflict and make sense."""
        prefix = self.arguments.prefix
        prefixv4 = self.arguments.prefixv4
        prefixv6 = self.arguments.prefixv6

        if prefix and (prefixv4 or prefixv6):
            exc_log = (
                "Specifying argument --prefix together with "
                "--prefixv4/--prefixv6 doesn't make sense"
            )
            raise RipeAtlasToolsException(exc_log)
        if prefix:
            return {"prefix": prefix}

        prefix_args = {}
        if prefixv4:
            prefix_args["prefix_v4"] = prefixv4

        if prefixv6:
            prefix_args["prefix_v6"] = prefixv6

        return prefix_args

    def _clean_location(self):
        """Make sure location argument are sane."""
        lat, lng = self.location2degrees()
        if self.arguments.radius:
            location_args = {
                "radius": "{0},{1}:{2}".format(lat, lng, self.arguments.radius)
            }
        else:
            location_args = {"latitude": lat, "longitude": lng}

        return location_args

    def location2degrees(self):
        """Fetches degrees based on the given location."""
        error_log = (
            "Following error occured while trying to fetch lat/lon"
            "for location <{}>:\n{}"
        )
        goole_api_url = "http://maps.googleapis.com/maps/api/geocode/json"
        try:
            result = requests.get(goole_api_url, params={
                "sensor": "false",
                "address": self.arguments.location
            })
        except (
            requests.ConnectionError,
            requests.HTTPError,
            requests.Timeout,
        ) as e:
            error_log = error_log.format(self.arguments.location, e)
            raise RipeAtlasToolsException(error_log)

        result = result.json()

        try:
            lat = result["results"][0]["geometry"]["location"]["lat"]
            lng = result["results"][0]["geometry"]["location"]["lng"]
        except (KeyError, IndexError) as e:
            error = error_log.format(self.arguments.location, e)
            raise RipeAtlasToolsException(error)

        return str(lat), str(lng)

    def _clean_center(self):
        """Make sure center argument are sane."""
        try:
            lat, lng = self.arguments.center.split(",")
        except ValueError:
            raise RipeAtlasToolsException(
                "Point argument should be in <lat,lng> format."
            )

        if self.arguments.radius:
            center_args = {
                "radius": "{0},{1}:{2}".format(lat, lng, self.arguments.radius)
            }
        else:
            center_args = {"latitude": lat, "longitude": lng}

        return center_args

    def _clean_country_code(self):
        """Make sure country_code argument are sane."""
        return {"country_code": self.arguments.country}

    def set_aggregators(self):
        """
        Builds and returns the key aggregators that will be used in
        the aggregation.
        """

        self.aggregators = []

        if not self.arguments.aggregate_by:
            return

        for key in self.arguments.aggregate_by:
            if key == "country":
                self.aggregators.append(ValueKeyAggregator(
                    key="country_code", prefix="Country"))
            else:
                self.aggregators.append(ValueKeyAggregator(key=key))

    def _get_line_items(self, probe):

        r = []

        for field in self.arguments.field:
            if field == "country":
                r.append((probe.country_code or "").lower())
            elif field in ("asn_v4", "asn_v6"):
                r.append(getattr(probe, field) or "")
            elif field == "description":
                description = sanitise(probe.description) or ""
                r.append(description[:self.COLUMNS["description"][1]])
            elif field == "coordinates":
                if probe.geometry and probe.geometry["coordinates"]:
                    lng, lat = probe.geometry["coordinates"]
                else:
                    lng, lat = UNK_COORDS
                r.append(u"{},{}".format(lat, lng))
            elif field in ("is_public", "is_anchor"):
                if getattr(probe, field):
                    r.append(u"\u2714")  # Check mark
                else:
                    r.append(u"\u2718")  # X
            else:
                r.append(sanitise(getattr(probe, field)))

        return r

    @staticmethod
    def _get_colour_from_status(status):
        if status == "Connected":
            return "green"
        if status == "Disconnected":
            return "yellow"
        if status == "Abandoned":
            return "red"
        return "white"

    def _get_line_format(self):
        r = TabularFieldsMixin._get_line_format(self)
        if not self.aggregators:
            return r
        return (u" " * len(self.aggregators)) + r

    def _get_header_names(self):
        r = []
        for field in self.arguments.field:
            if field == "id":
                r.append("ID")
            elif field == "is_public":
                r.append("Public")
            elif field == "is_anchor":
                r.append("Anchor")
            else:
                r.append(field.capitalize())
        return r

    def _get_line(self, probe):
        """
        Returns a utf8 encode string safe for printing and outputing
        to files.
        """
        log = colourise(
            self._get_line_format().format(*self._get_line_items(probe)),
            self._get_colour_from_status(probe.status)
        )

        return log

    def _get_filter_key_value_pair(self, k, v):
        if k == "country_code":
            return "Country", v.upper()
        if k == "asn":
            return "ASN", v
        return TabularFieldsMixin._get_filter_key_value_pair(self, k, v)



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from ripe.atlas.sagan import Result
from ripe.atlas.tools.renderers.http import Renderer


class TestHttpRenderer(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.basic = Result.get('{"lts":64,"from":"217.13.64.36","msm_id":2841267,"fw":4720,"timestamp":1450185727,"uri":"http://at-vie-as1120.anchors.atlas.ripe.net:80/4096","prb_id":1,"result":[{"rt":45.953289,"src_addr":"217.13.64.36","hsize":131,"af":4,"bsize":1668618,"res":200,"method":"GET","ver":"1.1","dst_addr":"193.171.255.2"}],"group_id":2841267,"type":"http","msm_name":"HTTPGet"}')
        self.multiple = Result.get('{"lts":64,"from":"217.13.64.36","msm_id":2841267,"fw":4720,"timestamp":1450185727,"uri":"http://at-vie-as1120.anchors.atlas.ripe.net:80/4096","prb_id":1,"result":[{"rt":45.953289,"src_addr":"217.13.64.36","hsize":131,"af":4,"bsize":1668618,"res":200,"method":"GET","ver":"1.1","dst_addr":"193.171.255.2"},{"rt":45.953289,"src_addr":"217.13.64.36","hsize":131,"af":4,"bsize":1668618,"res":200,"method":"GET","ver":"1.1","dst_addr":"193.171.255.2"}],"group_id":2841267,"type":"http","msm_name":"HTTPGet"}')

    def test_basic(self):
        expected = (
            '#Version: 1.0\n'
            '#Date: 2015-12-15 13:22:07\n'
            '#Fields: cs-method cs-uri c-ip s-ip sc-status time-taken http-version header-bytes body-bytes\n'
            'GET http://at-vie-as1120.anchors.atlas.ripe.net:80/4096 217.13.64.36 193.171.255.2 200 45.953289 1.1 131 1668618\n\n'
        )
        self.assertEqual(Renderer().on_result(self.basic), expected)

    def test_multiple(self):
        expected = (
            '#Version: 1.0\n'
            '#Date: 2015-12-15 13:22:07\n'
            '#Fields: cs-method cs-uri c-ip s-ip sc-status time-taken http-version header-bytes body-bytes\n'
            'GET http://at-vie-as1120.anchors.atlas.ripe.net:80/4096 217.13.64.36 193.171.255.2 200 45.953289 1.1 131 1668618\n'
            'GET http://at-vie-as1120.anchors.atlas.ripe.net:80/4096 217.13.64.36 193.171.255.2 200 45.953289 1.1 131 1668618\n\n'
        )
        self.assertEqual(Renderer().on_result(self.multiple), expected)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ripe.atlas.cousteau import Measurement
from ripe.atlas.cousteau.exceptions import APIResponseError

from .base import Command as BaseCommand, MetaDataMixin
from ..exceptions import RipeAtlasToolsException
from ..helpers.colours import colourise
from ..helpers.sanitisers import sanitise
from ..helpers.validators import ArgumentType


class Command(MetaDataMixin, BaseCommand):

    NAME = "measurement-info"
    DESCRIPTION = (
        "Return the meta data for one measurement"
    )

    def add_arguments(self):
        self.parser.add_argument("id", type=ArgumentType.msm_id_or_name(),
                                 help="The measurement id or alias")

    def run(self):

        try:
            measurement = Measurement(
                id=self.arguments.id, user_agent=self.user_agent)
        except APIResponseError:
            raise RipeAtlasToolsException(
                "That measurement does not appear to exist")

        self.render_basic(measurement)
        getattr(self, "render_{}".format(measurement.type.lower()))(measurement)

    @classmethod
    def render_basic(cls, measurement):
        url_template = "https://atlas.ripe.net/measurements/{}/"
        cls._render(measurement, (
            ("id", "ID"),
            ("id", "URL", lambda _: colourise(url_template.format(_), "cyan")),
            ("type", "Type", cls._prettify_type),
            ("status", "Status"),
            ("description", "Description", sanitise),
            ("af", "Address Family"),
            ("is_public", "Public?", cls._prettify_boolean),
            ("is_oneoff", "One-off?", cls._prettify_boolean),
            ("target", "Target Name", sanitise),
            ("target_address", "Target Address", sanitise),
            ("target_asn", "Target ASN"),
            ("interval", "Interval"),
            ("spread", "Spread"),
            ("creation_time", "Created", cls._prettify_time),
            ("start_time", "Started", cls._prettify_time),
            ("stop_time", "Stopped", cls._prettify_time),
            ("probes_requested", "Probes Requested"),
            ("probes_scheduled", "Probes Scheduled"),
            ("probes_currently_involved", "Probes Involved"),
            ("participant_count", "Participant Count"),
            ("is_all_scheduled", "Fully Scheduled?", cls._prettify_boolean),
            ("resolved_ips", "Resolved IPs", lambda _: ", ".join(_)),
            ("resolve_on_probe", "Resolve on the Probe", cls._prettify_boolean),
        ))

    @classmethod
    def render_ping(cls, measurement):
        cls._render(measurement, (
            ("packets", "Packets"),
            ("size", "Size"),
        ))

    @classmethod
    def render_traceroute(cls, measurement):
        cls._render(measurement, (
            ("packets", "Packets"),
            ("protocol", "Protocol"),
            ("dont_fragment", "Don't Fragment", cls._prettify_boolean),
            ("paris", "Paris"),
            ("first_hop", "First Hop"),
            ("max_hops", "Maximum Hops"),
            ("timeout", "Timeout"),
            ("size", "Size"),
            ("destination_option_size", "Destination Option Size"),
            ("hop_by_hop_option_size", "Hop-by-hop Option Size"),
            ("gap_limit", "Gap Limit"),
        ))

    @classmethod
    def render_dns(cls, measurement):
        cls._render(measurement, (
            ("query", "Query", cls._prettify_query),
            ("retry", "Retry Times"),
            ("include_qbuf", "Include the Qbuf?", cls._prettify_boolean),
            ("include_abuf", "Include the Abuf?", cls._prettify_boolean),
            ("protocol", "Protocol"),
            ("prepend_probe_id", "Prepend the Probe ID?"),
            ("udp_payload_size", "UDP Payload Size"),
            (
                "use_probe_resolver",
                "Use the Probe's Resolver?",
                cls._prettify_boolean
            ),
            ("set_do_bit", "Set the DO Bit?", cls._prettify_boolean),
            ("set_nsid_bit", "Set the NSID Bit?", cls._prettify_boolean),
            ("set_rd_bit", "Set the RD Bit?", cls._prettify_boolean),
            ("set_cd_bit", "Set the CD Bit?", cls._prettify_boolean),
        ))

    @classmethod
    def render_sslcert(cls, measurement):
        cls._render(measurement, (
            ("port", "Port"),
        ))

    @classmethod
    def render_http(cls, measurement):

        cls._render(measurement, (
            ("header_bytes", "Header Bytes"),
            ("version", "Version"),
            ("method", "Method"),
            ("port", "Port"),
            ("path", "Path", sanitise),
            ("query_string", "Query String", sanitise),
            ("user_agent", "User-Agent"),
            ("max_bytes_read", "Body Bytes"),
        ))

        timing_verbosity = 0
        if "extended_timing" in measurement.meta_data:
            if measurement.meta_data["extended_timing"]:
                timing_verbosity = 1
                if "more_extended_timing" in measurement.meta_data:
                    if measurement.meta_data["more_extended_timing"]:
                        timing_verbosity = 2
        cls._render_line("Timing Verbosity", timing_verbosity)

    @classmethod
    def render_ntp(cls, measurement):
        cls._render(measurement, (
            ("packets", "Packets"),
            ("timeout", "Timeout"),
        ))

    @staticmethod
    def _prettify_type(kind):
        types = {
            "ping": "Ping",
            "traceroute": "Traceroute",
            "dns": "DNS",
            "sslcert": "SSL Certificate",
            "http": "HTTP",
            "ntp": "NTP"
        }
        if kind in types:
            return colourise(colourise(types[kind], "bold"), "blue")
        return colourise("Unknown", "red")

    @staticmethod
    def _prettify_query(query):
        return sanitise(
            "{} {} {}".format(query["class"], query["type"], query["value"]))

    @classmethod
    def _render(cls, measurement, keys):
        for prop in keys:

            value = cls._get_measurement_property(measurement, prop[0])

            if value != "-" and len(prop) == 3:
                value = prop[2](value)

            cls._render_line(prop[1], value)

    @classmethod
    def _get_measurement_property(cls, measurement, property_name):
        value = getattr(measurement, property_name, None)

        if value is None and property_name in measurement.meta_data:
            value = measurement.meta_data[property_name]

        if value is None:
            value = "-"

        return value



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six

FORBIDDEN = dict((i, None) for i in list(range(0, 32)) + [127])


def sanitise(s, strip_newlines=True):
    """
    Strip out control characters to prevent people from screwing with the output
    """

    if not isinstance(s, six.string_types):
        return s

    if six.PY2:
        s = unicode(s)

    if not strip_newlines:
        return s.translate(
            dict((k, v) for k, v in FORBIDDEN.items() if not k == 10))

    return s.translate(FORBIDDEN)



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import functools
import os
import sys

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import anydbm as dbm  # anydbm py2.7 will use the best available dbm
except ImportError:
    import dbm  # ... and on Python3 dbm does the same


class LocalCache(object):
    """
    Simple caching engine, making use of the built-in dbm support.  This will
    create a file called cache.db in ripe-atlas-tools config directory and dump
    stuff in there for use later.
    """

    def __init__(self):
        self._now = datetime.datetime.now()
        self._db_file = None

    @property
    def _db(self):
        if self._db_file is None:
            self._db_file = dbm.open(self._get_or_create_db_path(), "c")
        return self._db_file

    def __contains__(self, key):
        return key in self._db

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value, expires=None):
        self._db[key] = pickle.dumps((expires, value))

    def __delitem__(self, key):
        if key not in self._db:
            raise KeyError
        del self._db[key]

    def keys(self):
        return self._db.keys()

    def items(self):
        for key in self.keys():
            yield key, self._db[key]

    def get(self, key, default=None):
        if key in self._db:
            expires, value = pickle.loads(self._db[key])
            if not expires or expires > self._now:
                return value
            else:
                del(self._db[key])
        return default

    def set(self, key, value, expires=None):
        return self.__setitem__(
            key, value, self._now + datetime.timedelta(seconds=expires))

    def clear(self, key=None):
        """
        Removes a specific key from the cache manually, or will wipe the entire
        cache if you don't specify `key`.  Note that this shouldn't be necessary
        unless you've cached something with an inappropriately long expire time.
        """
        if key:
            if key in self._db:
                del(self._db[key])
        else:
            for key in self.keys():
                del(self._db[key])

    def expire(self):
        """
        Clears out should-be-expired values from the cache.  Note that this
        happens automatically whenever you call `.get()` so you should never
        really need to run this.
        """
        for key in self.keys():
            self.get(key)

    @staticmethod
    def _get_or_create_db_path():

        v = sys.version_info
        file_name = "cache-{}.{}.{}.db".format(v.major, v.minor, v.micro)

        db_path = os.path.join("/", "tmp", file_name)
        if "HOME" in os.environ:
            db_path = os.path.join(
                os.environ["HOME"], ".config", "ripe-atlas-tools", file_name)

        try:
            os.makedirs(os.path.dirname(db_path))
        except OSError:
            pass  # Better to ask forgiveness than permission

        return db_path

cache = LocalCache()


class Memoiser(object):
    """
    Enabling class for the @memoised decorator
    """

    def __init__(self, function, cache_time):
        self._function = function
        self._cache_time = cache_time

    def __call__(self, *args, **kwargs):

        key = pickle.dumps([args, kwargs])
        value = cache[key]

        if value:
            return value

        value = self._function(*args, **kwargs)
        cache.set(key, value, self._cache_time)

        return value

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)


def memoised(cache_time):
    """
    Decorate a method or function with this to cache the result of said method
    for n seconds:

    @memoised(60 * 60)  # Cache for one hour
    my_function(some_arguments):
        ...
        return whatever
    """
    def _wrap(function):
        return Memoiser(function, cache_time=cache_time)
    return _wrap



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

import argparse
import os
import re
import sys

from dateutil import parser

from ..settings import aliases


class ArgumentType(object):

    @staticmethod
    def path(string):
        if not os.path.exists(string) and not string == "-":
            raise argparse.ArgumentTypeError(
                'The file name specified, "{}" does not appear to exist'.format(
                    string
                )
            )
        return string

    @staticmethod
    def country_code(string):
        if not re.match(r"^[a-zA-Z][a-zA-Z]$", string):
            raise argparse.ArgumentTypeError(
                "Countries must be defined with a two-letter ISO code")
        return string.upper()

    @staticmethod
    def datetime(string):
        try:
            return parser.parse(string)
        except:
            raise argparse.ArgumentTypeError(
                "Times must be specified in ISO 8601 format.  For example: "
                "2010-10-01T00:00:00 or a portion thereof.  All times are in "
                "UTC."
            )

    @staticmethod
    def ip_or_domain(string):
        message = '"{}" does not appear to be an IP address or host ' \
                  'name'.format(string)

        if " " in string:
            raise argparse.ArgumentTypeError(message)
        if "." not in string and ":" not in string:
            if not re.match(r"^\w+$", string):
                raise argparse.ArgumentTypeError(message)

        return string

    @classmethod
    def comma_separated_integers_or_file(cls, string):
        """
        Allow a list of comma-separated integers, or a file containing a
        newline-separated list of integers, OR "-" which implies standard out.
        """

        if re.match(r"^((\d+,?)+)$", string):
            return cls.comma_separated_integers()(string)

        f = sys.stdin
        if not string == "-":
            if not os.path.exists(string):
                raise argparse.ArgumentTypeError("Cannot find file: {}".format(
                    string
                ))
            f = open(string)

        try:
            return [int(_) for _ in f.readlines()]
        except ValueError:
            raise argparse.ArgumentTypeError(
                "The contents of the file presented does not conform to input "
                "standards.  Please ensure that every line in the file "
                "consists of a single integer."
            )

    @staticmethod
    def tag(string):
        pattern = re.compile("^[a-z_\-0-9]+$")

        if not pattern.match(string):
            raise argparse.ArgumentTypeError(
                '"{}" does not appear to be a valid tag.'.format(string))

        return string

    class integer_range(object):

        def __init__(self, minimum=float("-inf"), maximum=float("inf")):
            self.minimum = minimum
            self.maximum = maximum

        def __call__(self, string):

            message = "The integer must be between {} and {}.".format(
                self.minimum, self.maximum)
            if self.maximum == float("inf"):
                message = "The integer must be greater than {}.".format(
                    self.minimum)

            try:
                integer = int(string)
                if integer < self.minimum or integer > self.maximum:
                    raise argparse.ArgumentTypeError(message)
            except ValueError:
                raise argparse.ArgumentTypeError(
                    "An integer must be specified."
                )

            return integer

    class comma_separated_integers(object):

        def __init__(self, minimum=float("-inf"), maximum=float("inf")):
            self.minimum = minimum
            self.maximum = maximum

        def __call__(self, string):

            r = []

            for i in string.split(","):

                try:
                    i = int(i)
                except ValueError:
                    raise argparse.ArgumentTypeError(
                        "The ids supplied were not in the correct format. Note "
                        "that you must specify them as a list of "
                        "comma-separated integers without spaces.  Example: "
                        "1,2,34,157,10006"
                    )

                if i < self.minimum:
                    raise argparse.ArgumentTypeError(
                        "{} is lower than the minimum permitted value of "
                        "{}.".format(i, self.minimum)
                    )
                if i > self.maximum:
                    raise argparse.ArgumentTypeError(
                        "{} exceeds the maximum permitted value of {}.".format(
                            i, self.maximum)
                    )

                r.append(i)

            return r

    class regex(object):

        def __init__(self, regex):
            self.regex = re.compile(regex)

        def __call__(self, string):

            if not self.regex.match(string):
                raise argparse.ArgumentTypeError(
                    '"{}" does not appear to be valid.'.format(string))

            return string

    @staticmethod
    def alias_is_valid(string):
        ret = None

        if string and not string.isdigit():
            pattern = re.compile("^[a-zA-Z\._\-0-9]+$")

            if pattern.match(string):
                ret = string

        if not ret:
            raise argparse.ArgumentTypeError(
                '"{}" does not appear to be a valid '
                'alias.'.format(string))

        return ret

    class id_or_alias(object):
        TYPE = None

        def __call__(self, string):
            if string.isdigit():
                return int(string)

            if string in aliases[self.TYPE]:
                return int(aliases[self.TYPE][string])
            else:
                raise argparse.ArgumentTypeError(
                    '"{}" does not appear to be an existent '
                    '{} alias.'.format(string, self.TYPE)
                )

    class msm_id_or_name(id_or_alias):
        TYPE = "measurement"

    class probe_id_or_name(id_or_alias):
        TYPE = "probe"



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import requests
import IPy

from .cache import cache


class IP(object):

    RIPESTAT_URL = "https://stat.ripe.net/data/prefix-overview/data.json?resource={ip}"
    CACHE_EXPIRATION_TIME = 60 * 60 * 24 * 7

    def __init__(self, address):
        self.cached_prefix_found = False
        self.ip_object = IPy.IP(address)

        self.address = self.ip_object.strFullsize()
        self.asn = None
        self.holder = None
        self.prefix = None
        self.not_querable_types = [
            'RESERVED', 'UNSPECIFIED', 'LOOPBACK',
            'UNASSIGNED', 'DOCUMENTATION', 'ULA',
            'LINKLOCAL', 'PRIVATE'
        ]

        details = self._get_details()
        if details:
            self.asn = details["ASN"]
            self.holder = details["Holder"]
            self.prefix = details["Prefix"]

    def _get_details(self):
        details = None

        if not self.is_querable():
            return details

        details = cache.get("IPDetails:{}".format(self.address))
        if details:
            return details

        details = self.get_from_cached_prefix()

        if not details:
            details = self.query_stat()

        if details:
            self.update_cache(details)

        return details

    def is_querable(self):
        """Determines if address is worth querable."""
        return (self.ip_object.iptype() not in self.not_querable_types)

    def get_from_cached_prefix(self):
        """Search cache for existing cached Prefix"""
        details = None

        for cache_entry in cache.keys():
            if not cache_entry.decode().startswith("IPDetailsPrefix:"):
                continue

            prefix_details = cache.get(cache_entry)

            # data could exist but expired
            if not prefix_details:
                continue

            prefix = IPy.IP(prefix_details["Prefix"])

            if self.ip_object in prefix:
                details = prefix_details
                self.cached_prefix_found = True
                break

        return details

    def query_stat(self):
        """Query RIPE Stat to get address details."""
        URL = self.RIPESTAT_URL.format(ip=self.address)
        details = {}

        try:
            response = requests.get(URL)
            if not response.ok:
                return details
            res = response.json()
        except (requests.exceptions.RequestException, ValueError):
            # Catch all requests exception + not valid json ones.
            return details

        if res.get("status") == "ok":
            try:
                details = {
                    "ASN": str(res["data"]["asns"][0]["asn"]),
                    "Holder": res["data"]["asns"][0]["holder"],
                    "Prefix": res["data"]["resource"]
                }
            except (
                # Protect from any kind of malformed json response
                AttributeError, ValueError, KeyError, IndexError, TypeError
            ):
                pass

        return details

    def update_cache(self, details):
        """Update cache for the address and prefix if needed."""
        if not self.cached_prefix_found:
            key = "IPDetailsPrefix:{}".format(details["Prefix"])
            cache.set(key, details, self.CACHE_EXPIRATION_TIME)

        key = "IPDetails:{}".format(self.address)
        cache.set(key, details, self.CACHE_EXPIRATION_TIME)

    def __str__(self):
        return "IP {}, ASN {}, Holder {}".format(
            self.address, self.asn, self.holder
        )



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, absolute_import

from ...helpers.validators import ArgumentType
from ...settings import conf

from .base import Command


class HttpMeasureCommand(Command):

    def add_arguments(self):

        Command.add_arguments(self)

        spec = conf["specification"]["types"]["http"]

        specific = self.parser.add_argument_group("HTTP-specific Options")
        specific.add_argument(
            "--header-bytes",
            type=ArgumentType.integer_range(minimum=0, maximum=2048),
            default=spec["header-bytes"],
            help="The maximum number of bytes to retrieve from the header"
        )
        specific.add_argument(
            "--version",
            type=str,
            default=spec["version"],
            help="The HTTP version to use"
        )
        specific.add_argument(
            "--method",
            type=str,
            default=spec["method"],
            help="The HTTP method to use"
        )
        specific.add_argument(
            "--port",
            type=ArgumentType.integer_range(minimum=1, maximum=65535),
            default=spec["port"],
            help="Destination port"
        )
        specific.add_argument(
            "--path",
            type=str,
            default=spec["path"],
            help=""
        )
        specific.add_argument(
            "--query-string",
            type=str,
            default=spec["query-string"],
            help=""
        )
        specific.add_argument(
            "--user-agent",
            type=str,
            default=spec["user-agent"],
            help="The user agent used when performing the request"
        )
        specific.add_argument(
            "--body-bytes",
            type=ArgumentType.integer_range(minimum=1, maximum=1020048),
            default=spec["body-bytes"],
            help="The maximum number of bytes to retrieve from the body"
        )
        specific.add_argument(
            "--timing-verbosity",
            type=int,
            choices=(0, 1, 2),
            default=spec["timing-verbosity"],
            help="The amount of timing information you want returned. 1 "
                 "returns the time to read, to connect, and to first byte, 2 "
                 "returns timing information per read system call.  0 "
                 "(default) returns no additional timing information."
        )

    def _get_measurement_kwargs(self):

        r = Command._get_measurement_kwargs(self)

        keys = (
            "header_bytes", "version", "method", "port", "path", "query_string",
            "user_agent"
        )
        for key in keys:
            r[key] = getattr(self.arguments, key)

        if self.arguments.timing_verbosity > 0:
            r["extended_timing"] = True
            if self.arguments.timing_verbosity > 1:
                r["more_extended_timing"] = True

        r["max_bytes_read"] = self.arguments.body_bytes

        return r



# Copyright (c) 2016 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import webbrowser
from .base import Command as BaseCommand
from ..helpers.validators import ArgumentType


class Command(BaseCommand):

    NAME = "go"

    DESCRIPTION = "Visit the web page for a specific measurement"
    URL = "https://atlas.ripe.net/measurements/{0}/"

    def add_arguments(self):
        self.parser.add_argument(
            "measurement_id",
            type=ArgumentType.msm_id_or_name(),
            help="The measurement id or alias you want reported"
        )

    def run(self):
        url = self.URL.format(self.arguments.measurement_id)
        if not webbrowser.open(url):
            self.ok(
                "It looks like your system doesn't have a web browser "
                "available.  You'll have to go there manually: {0}".format(url)
            )



# Copyright (c) 2015 RIPE NCC
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from sphinx.application import Sphinx


class DocTest(unittest.TestCase):

    SOURCE_DIR = "docs"
    CONFIG_DIR = "docs"
    OUTPUT_DIR = "docs/build"
    DOCTREE_DIR = "docs/build/doctrees"

    def test_html_documentation(self):
        Sphinx(
            self.SOURCE_DIR,
            self.CONFIG_DIR,
            self.OUTPUT_DIR,
            self.DOCTREE_DIR,
            buildername="html",
            warningiserror=True,
        ).build(force_all=True)

    def test_text_documentation(self):
        Sphinx(
            self.SOURCE_DIR,
            self.CONFIG_DIR,
            self.OUTPUT_DIR,
            self.DOCTREE_DIR,
            buildername="text",
            warningiserror=True,
        ).build(force_all=True)
